# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import pytest

import hyperspy.api as hs
from hyperspy.signal_tools import ImageContrastEditor


class TestContrastEditorTool:

    def setup_method(self, method):
        s = hs.signals.Signal2D(np.arange(2*3*10*10).reshape(2, 3, 10, 10))
        self.s = s

    def test_reset_vmin_vmax(self):
        s = self.s
        s.plot(vmin='10th', vmax='99th')
        ceditor = ImageContrastEditor(s._plot.signal_plot)

        np.testing.assert_allclose(ceditor._vmin, 9.9)
        np.testing.assert_allclose(ceditor._vmax, 98.01)

        ceditor._vmin = 20
        ceditor._vmax = 90
        ceditor._reset_original_settings()

        np.testing.assert_allclose(ceditor._vmin, 9.9)
        np.testing.assert_allclose(ceditor._vmax, 98.01)

    def test_reset_span_selector(self):
        s = self.s
        s.plot(vmin='10th', vmax='99th')
        ceditor = ImageContrastEditor(s._plot.signal_plot)

        ceditor.span_selector.set_initial((20, 90))

        try:
            ceditor.update_span_selector()

            ax_image = s._plot.signal_plot.ax.images[0]

            np.testing.assert_allclose(ax_image.norm.vmin, 20)
            np.testing.assert_allclose(ax_image.norm.vmax, 90)

            ceditor._reset_span_selector()

            np.testing.assert_allclose(ax_image.norm.vmin, 9.9)
            np.testing.assert_allclose(ax_image.norm.vmax, 98.01)

        except TypeError as e:
            # Failure sometimes seen with pytest-xdist (parallel tests)
            # Message: `TypeError: restore_region() argument 1 must be matplotlib.backends._backend_agg.BufferRegion, not None`
            # See e.g. https://github.com/hyperspy/hyperspy/issues/1688
            # for a similar issue. Currently unclear what solution is.
            pytest.skip(f"Skipping reset_span_selector test due to {e}")


    def test_change_navigation_coordinate(self):
        s = self.s
        s.plot(vmin='10th', vmax='99th')
        ceditor = ImageContrastEditor(s._plot.signal_plot)

        np.testing.assert_allclose(ceditor._vmin, 9.9)
        np.testing.assert_allclose(ceditor._vmax, 98.01)
        try:
            # Convenience to be able to run test on systems using backends
            # supporting blit
            s.axes_manager.indices = (1, 1)
        except TypeError:
            pass

        np.testing.assert_allclose(ceditor._vmin, 409.9)
        np.testing.assert_allclose(ceditor._vmax, 498.01)

    def test_vmin_vmax_changed(self):
        s = self.s
        s.plot(vmin='0th', vmax='100th')
        ceditor = ImageContrastEditor(s._plot.signal_plot)

        np.testing.assert_allclose(ceditor._vmin, 0.0)
        np.testing.assert_allclose(ceditor._vmax, 99.0)
        try:
            # Convenience to be able to run test on systems using backends
            # supporting blit
            ceditor._vmin_percentile_changed(0, 10)
        except TypeError:
            pass
        try:
            # Convenience to be able to run test on systems using backends
            # supporting blit
            ceditor._vmax_percentile_changed(100, 99)
        except TypeError:
            pass

        np.testing.assert_allclose(ceditor._vmin, 9.9)
        np.testing.assert_allclose(ceditor._vmax, 98.01)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import inspect


def _connect_events(event, to_connect):
    try:
        for ev in event:
            # Iterable of events, connect all of them
            ev.connect(to_connect, [])
    except TypeError:
        # It was not an iterable, connect the single event
        event.connect(to_connect, [])


class Interactive:

    """Chainable operations on Signals that update on events.

    """

    def __init__(self, f, event="auto",
                 recompute_out_event="auto",
                 *args, **kwargs):
        """Update operation result when a given event is triggered.

        Parameters
        ----------
        f : function or method
            A function that returns an object and that optionally can place the
            result in an object given through the `out` keyword.
        event : {Event, "auto", None, iterable of events}
            Update the result of the operation when the event is triggered.
            If "auto" and `f` is a method of a Signal class instance its
            `data_changed` event is selected if the function takes an `out`
            argument. If None, `update` is not connected to any event. The
            default is "auto". It is also possible to pass an iterable of
            events, in which case all the events are connected.
        recompute_out_event : {Event, "auto", None, iterable of events}
            Optional argument. If supplied, this event causes a full
            recomputation of a new object. Both the data and axes of the new
            object are then copied over to the existing `out` object. Only
            useful for `Signal` or other objects that have an attribute
            `axes_manager`. If "auto" and `f` is a method of a Signal class
            instance its `AxesManager` `any_axis_chaged` event is selected.
            Otherwise the `Signal` `data_changed` event is selected.
            If None, `recompute_out` is not connected to any event.
            The default is "auto". It is also possible to pass an iterable of
            events, in which case all the events are connected.
        *args
            Arguments to be passed to `f`.
        **kwargs
            Keyword arguments to be passed to `f`.

        """
        from hyperspy.signal import BaseSignal
        self.f = f
        self.args = args
        self.kwargs = kwargs
        _plot_kwargs = self.kwargs.pop('_plot_kwargs', None)
        if 'out' in self.kwargs:
            self.f(*self.args, **self.kwargs)
            self.out = self.kwargs.pop('out')
        else:
            self.out = self.f(*self.args, **self.kwargs)
        # Reuse the `_plot_kwargs` for the roi if available
        if _plot_kwargs and 'signal' in self.kwargs:
            self.out._plot_kwargs = self.kwargs['signal']._plot_kwargs
        try:
            fargs = list(inspect.signature(self.f).parameters.keys())
        except TypeError:
            # This is probably a Cython function that is not supported by
            # inspect.
            fargs = []
        has_out = "out" in fargs
        # If it is a BaseSignal method
        if hasattr(f, "__self__") and isinstance(f.__self__, BaseSignal):
            if event == "auto":
                event = self.f.__self__.events.data_changed
            if recompute_out_event == "auto":
                recompute_out_event = \
                    self.f.__self__.axes_manager.events.any_axis_changed
        else:
            event = None if event == "auto" else event
            recompute_out_event = (None if recompute_out_event == "auto"
                                   else recompute_out_event)
        if recompute_out_event:
            _connect_events(recompute_out_event, self.recompute_out)
        if event:
            if has_out:
                _connect_events(event, self.update)
            else:
                #  We "simulate" out by triggering `recompute_out` instead.
                _connect_events(event, self.recompute_out)

    def recompute_out(self):
        out = self.f(*self.args, **self.kwargs)
        if out.data.shape == self.out.data.shape:
            # Keep the same array if possible.
            self.out.data[:] = out.data[:]
        else:
            self.out.data = out.data
        # The following may trigger an `any_axis_changed` event and, therefore,
        # it must precede the `data_changed` trigger below.
        self.out.axes_manager.update_axes_attributes_from(
            out.axes_manager._axes, attributes=["scale", "offset", "size"])
        self.out.events.data_changed.trigger(self.out)

    def update(self):
        self.f(*self.args, out=self.out, **self.kwargs)


def interactive(f, event="auto", recompute_out_event="auto", *args, **kwargs):
    cls = Interactive(f, event, recompute_out_event, *args, **kwargs)
    return cls.out


interactive.__doc__ = Interactive.__init__.__doc__



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import itertools

import numpy as np
import pytest

from hyperspy.components1d import Lorentzian
from hyperspy.signals import Signal1D
from hyperspy.utils import stack

TRUE_FALSE_2_TUPLE = [p for p in itertools.product((True, False), repeat=2)]


def test_function():
    g = Lorentzian()
    g.A.value = 1.5 * np.pi
    g.gamma.value = 1
    g.centre.value = 2
    np.testing.assert_allclose(g.function(2), 1.5)
    np.testing.assert_allclose(g.function(4), 0.3)


@pytest.mark.parametrize(("lazy"), (True, False))
@pytest.mark.parametrize(("only_current", "binned"), TRUE_FALSE_2_TUPLE)
def test_estimate_parameters_binned(only_current, binned, lazy):
    s = Signal1D(np.empty((250,)))
    s.metadata.Signal.binned = binned
    axis = s.axes_manager.signal_axes[0]
    axis.scale = .2
    axis.offset = -15
    g1 = Lorentzian(52342, 2, 10)
    s.data = g1.function(axis.axis)
    if lazy:
        s = s.as_lazy()
    g2 = Lorentzian()
    factor = axis.scale if binned else 1
    assert g2.estimate_parameters(s, axis.low_value, axis.high_value,
                                  only_current=only_current)
    assert g2.binned == binned
    np.testing.assert_allclose(g1.A.value, g2.A.value * factor,0.1)
    assert abs(g2.centre.value - g1.centre.value) <= 0.2
    assert abs(g2.gamma.value - g1.gamma.value) <= 0.1


@pytest.mark.parametrize(("lazy"), (True, False))
@pytest.mark.parametrize(("binned"), (True, False))
def test_function_nd(binned, lazy):
    s = Signal1D(np.empty((250,)))
    axis = s.axes_manager.signal_axes[0]
    axis.scale = .2
    axis.offset = -15
    g1 = Lorentzian(52342, 2, 10)
    s.data = g1.function(axis.axis)
    s.metadata.Signal.binned = binned
    s2 = stack([s] * 2)
    if lazy:
        s2 = s2.as_lazy()
    g2 = Lorentzian()
    factor = axis.scale if binned else 1
    g2.estimate_parameters(s2, axis.low_value, axis.high_value, False)
    assert g2.binned == binned
    np.testing.assert_allclose(g2.function_nd(axis.axis) * factor, s2.data,0.16)


def test_util_gamma_getset():
    g1 = Lorentzian()
    g1.gamma.value = 3.0
    np.testing.assert_allclose(g1.gamma.value, 3.0)


def test_util_fwhm_set():
    g1 = Lorentzian()
    g1.fwhm = 1.0
    np.testing.assert_allclose(g1.gamma.value, 0.5)


def test_util_fwhm_get():
    g1 = Lorentzian()
    g1.gamma.value = 2.0
    np.testing.assert_allclose(g1.fwhm, 4.0)


def test_util_fwhm_getset():
    g1 = Lorentzian()
    g1.fwhm = 4.0
    np.testing.assert_allclose(g1.fwhm, 4.0)


def test_util_height_set():
    g1 = Lorentzian()
    g1.gamma.value = 4.0
    g1.height = 2.0/np.pi
    np.testing.assert_allclose(g1.A.value, 8)


def test_util_height_get():
    g1 = Lorentzian()
    g1.gamma.value = 3.0
    g1.A.value = np.pi*1.5
    np.testing.assert_allclose(g1.height, 0.5)


def test_util_height_getset():
    g1 = Lorentzian()
    g1.height = 4.0
    np.testing.assert_allclose(g1.height, 4.0)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


from hyperspy._components.expression import Expression


class HeavisideStep(Expression):

    r"""The Heaviside step function.

    Based on the corresponding `numpy function
    <https://numpy.org/doc/stable/reference/generated/numpy.heaviside.html>`_
    using the half maximum definition for the central point:

    .. math::

        f(x) =
        \begin{cases}
        0 & x<n\\
        A/2 & x=n\\
        A & x>n
        \end{cases}


    ============== =============
    Variable        Parameter
    ============== =============
    :math:`n`       centre
    :math:`A`       height
    ============== =============


    Parameters
    -----------
    n : float
        Location parameter defining the x position of the step.
    A : float
        Height parameter for x>n.
    **kwargs
        Extra keyword arguments are passed to the ``Expression`` component.
    """

    def __init__(self, A=1., n=0., module="numpy", compute_gradients=True,
                 **kwargs):
        super(HeavisideStep, self).__init__(
            expression="A*heaviside(x-n,0.5)",
            name="HeavisideStep",
            A=A,
            n=n,
            position="n",
            module=module,
            autodoc=False,
            compute_gradients=compute_gradients,
            **kwargs)

        self.isbackground = True
        self.convolved = False




# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>


from collections import OrderedDict
import math as math
import logging

import numpy as np
from numba import njit

from hyperspy.misc.math_tools import anyfloatin


_logger = logging.getLogger(__name__)


def get_array_memory_size_in_GiB(shape, dtype):
    """Given the size and dtype returns the amount of memory that such
    an array needs to allocate

    Parameters
    ----------
    shape: tuple
    dtype : data-type
        The desired data-type for the array.
    """
    if not isinstance(dtype, np.dtype):
        dtype = np.dtype(dtype)
    return np.array(shape).cumprod()[-1] * dtype.itemsize / 2.0 ** 30


def are_aligned(shape1, shape2):
    """Check if two numpy arrays are aligned.

    Parameters
    ----------
    shape1, shape2 : iterable

    Returns
    -------
    isaligned : bool

    """
    isaligned = True
    shape1 = list(shape1)
    shape2 = list(shape2)
    try:
        while isaligned is True:
            dim1 = shape1.pop()
            dim2 = shape2.pop()
            if dim1 != dim2 and (dim1 != 1 and dim2 != 1):
                isaligned = False
    except IndexError:
        return isaligned
    return isaligned


def homogenize_ndim(*args):
    """Given any number of arrays returns the same arrays
    reshaped by adding facing dimensions of size 1.

    """

    max_len = max([len(ary.shape) for ary in args])

    return [ary.reshape((1,) * (max_len - len(ary.shape)) + ary.shape)
            for ary in args]


def _requires_linear_rebin(arr, scale):
    """Returns True if linear_rebin is required.
    Parameters
    ----------
    arr: array
        numpy array to rebin
    scale: tuple
        rebinning factors
    """

    return (np.asarray(arr.shape) %
            np.asarray(scale)).any() or anyfloatin(scale)


def rebin(a, new_shape=None, scale=None, crop=True):
    """Rebin array.

    rebin ndarray data into a smaller or larger array based on a linear
    interpolation. Specify either a new_shape or a scale. Scale of 1== no
    binning. Scale less than one results in up-sampling.

    Parameters
    ----------
    a : numpy array
    new_shape : a list of floats or integer, default None
        For each dimension specify the new_shape of the np.array. This will
        then be converted into a scale.
    scale : a list of floats or integer, default None
        For each dimension specify the new:old pixel ratio, e.g. a ratio of 1
        is no binning and a ratio of 2 means that each pixel in the new
        spectrum is twice the size of the pixels in the old spectrum.
        The length of the list should match the dimension of the numpy array.
        ***Note : Only one of scale or new_shape should be specified otherwise
        the function will not run***
    crop: bool, default True
        When binning by a non-integer number of pixels it is likely that
        the final row in each dimension contains less than the full quota to
        fill one pixel.

        e.g. 5*5 array binned by 2.1 will produce two rows containing 2.1
        pixels and one row containing only 0.8 pixels worth. Selection of
        crop='True' or crop='False' determines whether or not this
        'black' line is cropped from the final binned array or not.

        *Please note that if crop=False is used, the final row in each
        dimension may appear black, if a fractional number of pixels are left
        over. It can be removed but has been left to preserve total counts
        before and after binning.*

    Returns
    -------
    numpy array

    Examples
    --------
    >>> a=rand(6,4); b=rebin(a,scale=(3,2))
    >>> a=rand(6); b=rebin(a,scale=(2,))

    Notes
    -----
    Fast re_bin function Adapted from scipy cookbook
    If rebin function fails with error stating that the function is 'not binned
    and therefore cannot be rebinned', add binned to metadata with:
    >>> s.metadata.Signal.binned = True

    """
    # Series of if statements to check that only one out of new_shape or scale
    # has been given. New_shape is then converted to scale. If both or neither
    # are given the function raises and error and wont run.
    if new_shape is None and scale is None:
        raise ValueError("One of new_shape, or scale must be specified")
    elif new_shape is not None and scale is not None:
        raise ValueError("Only one out of new_shape or scale should be specified")
    elif new_shape is not None:
        scale = []
        for i, _ in enumerate(a.shape):
            scale.append(a.shape[i] / new_shape[i])
    else:
        new_shape = new_shape
        scale = scale
    # check whether or not interpolation is needed.
    if _requires_linear_rebin(arr=a, scale=scale):
        _logger.debug("Using linear_bin")
        if np.issubdtype(a.dtype, np.integer):
            # The _linear_bin function below requires a float dtype
            # because of the default numpy casting rule ('same_kind').
            a = a.astype("float", casting="safe", copy=False)
        return _linear_bin(a, scale, crop)
    else:
        _logger.debug("Using standard rebin with lazy support")
        # if interpolation is not needed run fast re_bin function.
        # Adapted from scipy cookbook.
        lenShape = len(a.shape)
        new_shape = np.asarray(a.shape) // np.asarray(scale)
        # ensure the new shape is integers
        new_shape = tuple(int(ns) for ns in new_shape)
        # check function wont bin to zero.
        for item in new_shape:
            if item == 0:
                raise ValueError(
                    "One of your dimensions collapses to zero. "
                    "Re-adjust your scale values or run code with "
                    "crop=False to avoid this error."
                )
        scale = np.asarray(a.shape) // np.asarray(new_shape)
        if scale.max() < 2:
            return a.copy()
        if isinstance(a, np.ndarray):
            # most of the operations will fall here and dask is not imported
            rshape = ()
            for athing in zip(new_shape, scale):
                rshape += athing
            return a.reshape(rshape).sum(axis=tuple(
                2 * i + 1 for i in range(lenShape)))
        else:
            import dask.array as da

            try:
                return da.coarsen(np.sum, a, {i: int(f)
                                              for i, f in enumerate(scale)})
            # we provide slightly better error message in hyperspy context
            except ValueError:
                raise ValueError(
                    "Rebinning does not align with data dask chunks. "
                    "Rebin fewer dimensions at a time to avoid this error"
                )


@njit(cache=True)
def _linear_bin_loop(result, data, scale):  # pragma: no cover
    for j in range(result.shape[0]):
        # Begin by determining the upper and lower limits of a given new pixel.
        x1 = j * scale
        x2 = min((1 + j) * scale, data.shape[0])
        value = result[j:j + 1]

        if (x2 - x1) >= 1:
            # When binning, the first part is to deal with the fractional pixel
            # left over from it being non-integer binning e.g. when x1=1.4
            cx1 = math.ceil(x1)
            rem = cx1 - x1
            # This will add a value of fractional pixel to the bin, eg if x1=1.4,
            # the fist step will be to add 0.6*data[1]
            value += data[math.floor(x1)] * rem
            # Update x1 to remove the part of the bin we have just added.
            x1 = cx1
            while (x2 - x1) >= 1:
                # Main binning function to add full pixel values to the data.
                value += data[int(x1)]
                # Update x1 each time.
                x1 += 1
            if x2 > x1:
                # Finally take into account the fractional pixel left over.
                value += data[math.floor(x1)] * (x2 - x1)
        else:
            # When step < 1, so we are upsampling
            fx1 = math.floor(x1)
            cx1 = math.ceil(x1)
            if scale > (cx1 - x1) > 0:
                # If our step is smaller than rounding up to the nearest whole
                # number.
                value += data[fx1] * (cx1 - x1)
                x1 = cx1  # This step is needed when this particular bin straddes
                # two neighbouring pixels.
            if x1 < x2:
                # The standard upsampling function where each new pixel is a
                # fraction of the original pixel.
                value += data[math.floor(x1)] * (x2 - x1)


def _linear_bin(dat, scale, crop=True):
    """Binning of the spectrum image by a non-integer pixel value.

    Parameters
    ----------
    originalSpectrum : numpy.array
    scale : a list of floats
        For each dimension specify the new:old pixel ratio,
        e.g. a ratio of 1 is no binning; a ratio of 2 means that each pixel in
        the new spectrum is twice the size of the pixels in the old spectrum.
        The length of the list should match the dimensions of the data.
    crop : bool, default True
        When binning by a non-integer number of pixels it is likely that
        the final row in each dimension contains less than the full quota to
        fill one pixel.
        e.g. 5*5 array binned by 2.1 will produce two rows containing 2.1
        pixels and one row containing only 0.8 pixels worth. Selection of
        crop='True' or crop='False' determines whether or not this 'black'
        line is cropped from the final binned array or not.

        *Please note that if crop=False is used, the final row in each
        dimension may appear black, if a fractional number of pixels are left
        over. It can be removed but has been left optional to preserve total
        counts before and after binning.*

    Returns
    -------
    np.array
        with new dimensions width/scale for each dimension in the data.
    """
    if len(dat.shape) != len(scale):
        raise ValueError(
            "The list of bins must match the number of dimensions, including "
            "the energy dimension. In order to not bin in any of these "
            "dimensions specifically, simply set the value in shape to 1."
        )

    for axis, s in enumerate(scale):
        # For each iteration of linear_bin the axis being interated over has to
        # be switched to axis[0] in order to carry out the interation loop.
        dat = np.swapaxes(dat, 0, axis)
        # The new dimension size is old_size/step, this is rounded down normally
        # but if crop is switched off it is rounded up to the nearest whole
        # number.
        if not np.issubdtype(s, np.floating):
            s = float(s)

        dim = (math.floor(dat.shape[0] / s) if crop
               else math.ceil(dat.shape[0] / s))
        # check function wont bin to zero.
        if dim == 0:
            raise ValueError(
                "One of your dimensions collapses to zero. "
                "Re-adjust your scale values or run code with "
                "crop=False to avoid this error."
            )
        # Set up the result np.array to have a new axis[0] size for after
        # cropping.
        result = np.zeros((dim,) + dat.shape[1:])
        # Carry out binning over axis[0]
        _linear_bin_loop(result=result, data=dat, scale=s)
        # Swap axis[0] back to the original axis location.
        result = result.swapaxes(0, axis)
        # Update the np.array reading of iterating over the next axis.
        dat = result

    return result


def sarray2dict(sarray, dictionary=None):
    """Converts a struct array to an ordered dictionary

    Parameters
    ----------
    sarray: struct array
    dictionary: None or dict
        If dictionary is not None the content of sarray will be appended to the
        given dictonary

    Returns
    -------
    Ordered dictionary

    """
    if dictionary is None:
        dictionary = OrderedDict()
    for name in sarray.dtype.names:
        dictionary[name] = sarray[name][0] if len(sarray[name]) == 1 else sarray[name]
    return dictionary


def dict2sarray(dictionary, sarray=None, dtype=None):
    """Populates a struct array from a dictionary

    Parameters
    ----------
    dictionary: dict
    sarray: struct array or None
        Either sarray or dtype must be given. If sarray is given, it is
        populated from the dictionary.
    dtype: None, numpy dtype or dtype list
        If sarray is None, dtype must be given. If so, a new struct array
        is created according to the dtype, which is then populated.

    Returns
    -------
    Structure array

    """
    if sarray is None:
        if dtype is None:
            raise ValueError("Either sarray or dtype need to be specified.")
        sarray = np.zeros((1,), dtype=dtype)
    for name in set(sarray.dtype.names).intersection(set(dictionary.keys())):
        if len(sarray[name]) == 1:
            sarray[name][0] = dictionary[name]
        else:
            sarray[name] = dictionary[name]
    return sarray


@njit(cache=True)
def numba_histogram(data, bins, ranges):
    """
    Parameters
    ----------
    data : numpy array
        Input data. The histogram is computed over the flattened array.
    bins : int
        Number of bins
    ranges : (float, float)
        The lower and upper range of the bins.

    Returns
    -------
    hist : array
        The values of the histogram.
    """
    # Adapted from https://iscinumpy.gitlab.io/post/histogram-speeds-in-python/
    hist = np.zeros((bins,), dtype=np.intp)
    delta = 1 / ((ranges[1] - ranges[0]) / bins)

    for x in data.flat:
        i = (x - ranges[0]) * delta
        if 0 <= i < bins:
            hist[int(i)] += 1

    return hist



# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy.misc.utils import DictionaryTreeBrowser
from hyperspy.samfire_utils.weights.red_chisq import ReducedChiSquaredWeight


class Test_Red_chisq_weight:

    def setup_method(self, method):
        self.w = ReducedChiSquaredWeight()
        artificial_model = DictionaryTreeBrowser()
        artificial_model.add_node('red_chisq.data')
        artificial_model.red_chisq.data = np.arange(35).reshape((5, 7))
        self.w.model = artificial_model

    def test_function(self):
        w = self.w
        ind = (2, 3)
        assert w.function(ind) == 16

    def test_map_noslice(self):
        w = self.w
        mask = np.ones((5, 7), dtype=bool)
        mask[0, 0] = False
        ans = w.map(mask)
        assert np.all(w.model.red_chisq.data[mask] - 1 == ans[mask])
        assert np.isnan(ans[0, 0])



# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from distutils.version import LooseVersion

import matplotlib
import pytest

from hyperspy.misc.test_utils import check_running_tests_in_CI


def test_mlp_agg_for_CI_testing():
    if check_running_tests_in_CI():
        assert matplotlib.get_backend() == 'agg'

@pytest.fixture
def mpl_cmdopt(request):
    return request.config.getoption("--mpl")

def test_mpl_version():
    # for simplicity, only matplotlib 2.x is supported for testing
    assert LooseVersion(matplotlib.__version__) >= LooseVersion('2.0.0')


@pytest.mark.xfail(reason="Check if plotting tests are working: if this test passes,"
                   " it means that the image comparison of the plotting test are"
                   " not working.",
                   strict=True)
@pytest.mark.mpl_image_compare(baseline_dir='', tolerance=2)
def test_plotting_test_working(mpl_cmdopt):
    # Skip if --mpl command line option is not present, because it will always
    # pass the image comparison test and, therefore, this test will always fail.
    if not mpl_cmdopt:
        pytest.skip("No --mpl command line option")
    # If this test passes, it means that the plotting tests are not working!
    # In this case, the test will be reported as failed because the xfail is
    # 'strict'.
    #
    # To check if the plotting test are working, we compare this plot with an
    # incorrect baseline image, so it would fail as expected.
    fig = matplotlib.pyplot.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([1, 2, 2])
    # to generate a different plot uncomment the next line
    # ax.plot([1, 2, 3, 4]) # Uncomment this line to make sure the test is
    # properly failing
    return fig



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from hyperspy._signals.signal1d import Signal1D
from hyperspy._signals.signal2d import Signal2D
from hyperspy.datasets import artificial_data
from hyperspy.decorators import lazifyTestClass
from hyperspy.misc.machine_learning.import_sklearn import sklearn_installed
from hyperspy.misc.machine_learning.tools import amari
from hyperspy.signals import BaseSignal


def are_bss_components_equivalent(c1_list, c2_list, atol=1e-4):
    """Check if two list of components are equivalent.

    To be equivalent they must differ by a max of `atol` except
    for an arbitrary -1 factor.

    Parameters
    ----------
    c1_list, c2_list: list of Signal instances.
        The components to check.
    atol: float
        Absolute tolerance for the amount that they can differ.

    Returns
    -------
    bool

    """
    matches = 0
    for c1 in c1_list:
        for c2 in c2_list:
            if np.allclose(c2.data, c1.data, atol=atol) or np.allclose(
                c2.data, -c1.data, atol=atol
            ):
                matches += 1
    return matches == len(c1_list)


def test_amari_distance(n=16, tol=1e-6):
    """Amari distance between matrix and its inverse should be 0."""
    rng = np.random.RandomState(123)

    A = rng.randn(n, n)
    W = np.linalg.inv(A)
    X = np.linalg.pinv(A)

    np.testing.assert_allclose(amari(W, A), 0.0, rtol=tol)
    np.testing.assert_allclose(amari(A, A), 2.912362, rtol=tol)
    np.testing.assert_allclose(amari(X, A), 0.0, rtol=tol)


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
def test_bss_FastICA_object():
    """Tests that a simple sklearn pipeline is an acceptable algorithm."""
    rng = np.random.RandomState(123)
    S = rng.laplace(size=(3, 1000))
    A = rng.random_sample(size=(3, 3))
    s = Signal1D(A @ S)
    s.decomposition()

    from sklearn.decomposition import FastICA

    out = s.blind_source_separation(
        3, algorithm=FastICA(algorithm="deflation"), return_info=True
    )

    assert hasattr(out, "components_")


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
def test_bss_pipeline():
    """Tests that a simple sklearn pipeline is an acceptable algorithm."""
    rng = np.random.RandomState(123)
    S = rng.laplace(size=(3, 1000))
    A = rng.random_sample(size=(3, 3))
    s = Signal1D(A @ S)
    s.decomposition()

    from sklearn.decomposition import FastICA
    from sklearn.pipeline import Pipeline
    from sklearn.preprocessing import StandardScaler

    est = Pipeline(
        [("scaler", StandardScaler()), ("ica", FastICA(algorithm="deflation"))]
    )
    out = s.blind_source_separation(3, algorithm=est, return_info=True)

    assert hasattr(out, "steps")
    assert hasattr(out.named_steps["ica"], "components_")


@pytest.mark.parametrize("whiten_method", ["PCA", "ZCA"])
def test_orthomax(whiten_method):
    rng = np.random.RandomState(123)
    S = rng.laplace(size=(3, 500))
    A = rng.random_sample(size=(3, 3))
    s = Signal1D(A @ S)
    s.decomposition()
    s.blind_source_separation(3, algorithm="orthomax", whiten_method=whiten_method)

    W = s.learning_results.unmixing_matrix
    assert amari(W, A) < 0.5

    # Verify that we can change gamma for orthomax method
    s = artificial_data.get_core_loss_eels_line_scan_signal()
    s.decomposition()
    s.blind_source_separation(2, algorithm="orthomax", gamma=2)


def test_no_decomposition_error():
    s = artificial_data.get_core_loss_eels_line_scan_signal()

    with pytest.raises(AttributeError, match="A decomposition must be performed"):
        s.blind_source_separation(2)


def test_factors_error():
    s = artificial_data.get_core_loss_eels_line_scan_signal()
    s.decomposition()

    factors = s.get_decomposition_factors().data

    with pytest.raises(TypeError, match="`factors` must be a BaseSignal instance"):
        s.blind_source_separation(2, factors=factors)

    factors = BaseSignal(s.get_decomposition_factors().data)

    with pytest.raises(ValueError, match="`factors` must have navigation dimension"):
        s.blind_source_separation(2, factors=factors)


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
@pytest.mark.parametrize("num_components", [None, 2])
def test_num_components(num_components):
    s = artificial_data.get_core_loss_eels_line_scan_signal()
    s.decomposition(output_dimension=2)
    s.blind_source_separation(number_of_components=num_components)


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
def test_components_list():
    s = artificial_data.get_core_loss_eels_line_scan_signal()
    s.decomposition(output_dimension=3)
    s.blind_source_separation(comp_list=[0, 2])
    assert s.learning_results.unmixing_matrix.shape == (2, 2)


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
def test_num_components_error():
    s = artificial_data.get_core_loss_eels_line_scan_signal()
    s.decomposition()
    s.learning_results.output_dimension = None

    with pytest.raises(
        ValueError, match="No `number_of_components` or `comp_list` provided"
    ):
        s.blind_source_separation(number_of_components=None)


def test_algorithm_error():
    s = artificial_data.get_core_loss_eels_line_scan_signal()
    s.decomposition()

    with pytest.raises(ValueError, match="'algorithm' not recognised"):
        s.blind_source_separation(2, algorithm="uniform")


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
def test_normalize_components_errors():
    rng = np.random.RandomState(123)
    s = Signal1D(rng.random_sample(size=(20, 100)))
    s.decomposition()

    with pytest.raises(ValueError, match="called after s.blind_source_separation"):
        s.normalize_bss_components(target="loadings")

    s.blind_source_separation(2)

    with pytest.raises(ValueError, match="target must be"):
        s.normalize_bss_components(target="uniform")


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
def test_sklearn_convergence_warning():
    # Import here to avoid error if sklearn missing
    from sklearn.exceptions import ConvergenceWarning

    rng = np.random.RandomState(123)
    ics = rng.laplace(size=(3, 1000))
    mixing_matrix = rng.random_sample(size=(100, 3))
    s = Signal1D(mixing_matrix @ ics)
    s.decomposition()

    with pytest.warns(ConvergenceWarning):
        s.blind_source_separation(
            number_of_components=10,
            algorithm="sklearn_fastica",
            diff_order=1,
            on_loadings=True,
            tol=1e-15,
            max_iter=3,
        )


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
@pytest.mark.parametrize("whiten_method", [None, "PCA", "ZCA"])
def test_fastica_whiten_method(whiten_method):
    rng = np.random.RandomState(123)
    S = rng.laplace(size=(3, 1000))
    A = rng.random_sample(size=(3, 3))
    s = Signal1D(A @ S)
    s.decomposition()
    s.blind_source_separation(
        3, algorithm="sklearn_fastica", whiten_method=whiten_method
    )
    assert s.learning_results.unmixing_matrix.shape == A.shape


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
@lazifyTestClass
class TestReverseBSS:
    def setup_method(self, method):
        rng = np.random.RandomState(123)
        S = rng.laplace(size=(3, 500))
        S -= 2 * S.min()  # Required to give us a positive dataset
        A = rng.random_sample(size=(3, 3))
        s = Signal1D(A @ S)
        s.decomposition()
        s.blind_source_separation(2)
        self.s = s

    def test_autoreverse_default(self):
        self.s.learning_results.bss_factors[:, 0] *= -1
        self.s._auto_reverse_bss_component("loadings")
        np.testing.assert_array_less(self.s.learning_results.bss_factors[:, 0], 0)
        np.testing.assert_array_less(0, self.s.learning_results.bss_factors[:, 1])
        self.s._auto_reverse_bss_component("factors")
        np.testing.assert_array_less(0, self.s.learning_results.bss_factors)

    def test_autoreverse_on_loading(self):
        self.s._auto_reverse_bss_component("loadings")
        np.testing.assert_array_less(0, self.s.learning_results.bss_factors)

    def test_reverse_wrong_parameter(self):
        with pytest.raises(ValueError):
            self.s.blind_source_separation(2, reverse_component_criterion="toto")


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
@lazifyTestClass
class TestBSS1D:
    def setup_method(self, method):
        rng = np.random.RandomState(123)
        ics = rng.laplace(size=(3, 500))
        mixing_matrix = rng.random_sample(size=(100, 3))
        s = Signal1D(mixing_matrix @ ics)
        s.decomposition()

        mask_sig = s._get_signal_signal(dtype="bool")
        mask_sig.isig[5] = True

        mask_nav = s._get_navigation_signal(dtype="bool")
        mask_nav.isig[5] = True

        self.s = s
        self.mask_nav = mask_nav
        self.mask_sig = mask_sig

    def test_on_loadings(self):
        self.s.blind_source_separation(3, diff_order=0, fun="exp", on_loadings=False)
        s2 = self.s.as_signal1D(0)
        s2.decomposition()
        s2.blind_source_separation(3, diff_order=0, fun="exp", on_loadings=True)
        assert are_bss_components_equivalent(
            self.s.get_bss_factors(), s2.get_bss_loadings()
        )

    @pytest.mark.filterwarnings("ignore:FastICA did not converge")
    @pytest.mark.parametrize("on_loadings", [True, False])
    @pytest.mark.parametrize("diff_order", [0, 1])
    def test_mask_diff_order(self, on_loadings, diff_order):
        if on_loadings:
            mask = self.mask_nav
            self.s.learning_results.loadings[5, :] = np.nan
        else:
            mask = self.mask_sig
            self.s.learning_results.factors[5, :] = np.nan

        # This preserves the old test behaviour. Without it,
        # we get "ConvergenceWarning: FastICA did not converge."
        # We test for convergence warnings separately
        n_components = 2 if diff_order == 1 and on_loadings else 3

        self.s.blind_source_separation(
            n_components, diff_order=diff_order, mask=mask, on_loadings=on_loadings
        )


@pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
@lazifyTestClass
class TestBSS2D:
    def setup_method(self, method):
        rng = np.random.RandomState(123)
        ics = rng.laplace(size=(3, 256))
        mixing_matrix = rng.random_sample(size=(100, 3))
        s = Signal2D((mixing_matrix @ ics).reshape((100, 16, 16)))
        for (axis, name) in zip(s.axes_manager._axes, ("z", "y", "x")):
            axis.name = name
        s.decomposition()

        mask_sig = s._get_signal_signal(dtype="bool")
        mask_sig.unfold()
        mask_sig.isig[5] = True
        mask_sig.fold()

        mask_nav = s._get_navigation_signal(dtype="bool")
        mask_nav.unfold()
        mask_nav.isig[5] = True
        mask_nav.fold()

        self.s = s
        self.mask_nav = mask_nav
        self.mask_sig = mask_sig

    def test_diff_axes_string_with_mask(self):
        self.s.learning_results.factors[5, :] = np.nan
        factors = self.s.get_decomposition_factors().inav[:3]
        if self.mask_sig._lazy:
            self.mask_sig.compute()
        self.s.blind_source_separation(
            3,
            diff_order=0,
            fun="exp",
            on_loadings=False,
            factors=factors.diff(axis="x", order=1),
            mask=self.mask_sig.diff(axis="x", order=1),
        )
        matrix = self.s.learning_results.unmixing_matrix.copy()
        self.s.blind_source_separation(
            3,
            diff_order=1,
            fun="exp",
            on_loadings=False,
            diff_axes=["x"],
            mask=self.mask_sig,
        )
        np.testing.assert_allclose(
            matrix, self.s.learning_results.unmixing_matrix, atol=1e-5
        )

    def test_diff_axes_string_without_mask(self):
        factors = self.s.get_decomposition_factors().inav[:3].diff(axis="x", order=1)
        self.s.blind_source_separation(
            3, diff_order=0, fun="exp", on_loadings=False, factors=factors
        )
        matrix = self.s.learning_results.unmixing_matrix.copy()
        self.s.blind_source_separation(
            3, diff_order=1, fun="exp", on_loadings=False, diff_axes=["x"],
        )
        np.testing.assert_allclose(
            matrix, self.s.learning_results.unmixing_matrix, atol=1e-3
        )

    def test_diff_axes_without_mask(self):
        factors = self.s.get_decomposition_factors().inav[:3].diff(axis="y", order=1)
        self.s.blind_source_separation(
            3, diff_order=0, fun="exp", on_loadings=False, factors=factors
        )
        matrix = self.s.learning_results.unmixing_matrix.copy()
        self.s.blind_source_separation(
            3, diff_order=1, fun="exp", on_loadings=False, diff_axes=[2],
        )
        np.testing.assert_allclose(
            matrix, self.s.learning_results.unmixing_matrix, atol=1e-3
        )

    def test_on_loadings(self):
        self.s.blind_source_separation(3, diff_order=0, fun="exp", on_loadings=False)
        s2 = self.s.as_signal1D(0)
        s2.decomposition()
        s2.blind_source_separation(3, diff_order=0, fun="exp", on_loadings=True)
        assert are_bss_components_equivalent(
            self.s.get_bss_factors(), s2.get_bss_loadings()
        )

    @pytest.mark.parametrize("diff_order", [0, 1])
    def test_mask_diff_order_0(self, diff_order):
        self.s.learning_results.factors[5, :] = np.nan
        self.s.blind_source_separation(3, diff_order=diff_order, mask=self.mask_sig)

    def test_mask_diff_order_1_diff_axes(self):
        self.s.learning_results.factors[5, :] = np.nan
        self.s.blind_source_separation(
            3, diff_order=1, mask=self.mask_sig, diff_axes=["x"]
        )

    def test_mask_diff_order_0_on_loadings(self):
        self.s.learning_results.loadings[5, :] = np.nan
        self.s.blind_source_separation(
            3, diff_order=0, mask=self.mask_nav, on_loadings=True
        )

    def test_mask_diff_order_1_on_loadings(self):
        s = self.s.to_signal1D()
        s.decomposition()
        if hasattr(s.learning_results.loadings, "compute"):
            s.learning_results.loadings = s.learning_results.loadings.compute()
        s.learning_results.loadings[5, :] = np.nan
        s.blind_source_separation(3, diff_order=1, mask=self.mask_sig, on_loadings=True)

    def test_mask_diff_order_1_on_loadings_diff_axes(self):
        s = self.s.to_signal1D()
        s.decomposition()
        if hasattr(s.learning_results.loadings, "compute"):
            s.learning_results.loadings = s.learning_results.loadings.compute()
        s.learning_results.loadings[5, :] = np.nan
        s.blind_source_separation(
            3, diff_order=1, mask=self.mask_sig, on_loadings=True, diff_axes=["x"]
        )


class TestPrintInfo:
    def setup_method(self, method):
        rng = np.random.RandomState(123)
        self.s = Signal1D(rng.random_sample(size=(20, 100)))
        self.s.decomposition(output_dimension=2)

    def test_bss(self, capfd):
        self.s.blind_source_separation(2, algorithm="orthomax")
        captured = capfd.readouterr()
        assert "Blind source separation info:" in captured.out

    @pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
    def test_bss_sklearn(self, capfd):
        self.s.blind_source_separation(2)
        captured = capfd.readouterr()
        assert "Blind source separation info:" in captured.out
        assert "scikit-learn estimator:" in captured.out


class TestReturnInfo:
    def setup_method(self, method):
        rng = np.random.RandomState(123)
        self.s = Signal1D(rng.random_sample(size=(20, 100)))
        self.s.decomposition(output_dimension=2)

    def test_bss_not_supported(self):
        assert (
            self.s.blind_source_separation(algorithm="orthomax", return_info=True)
            is None
        )

    @pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
    def test_bss_supported_return_true(self):
        assert self.s.blind_source_separation(return_info=True) is not None

    @pytest.mark.skipif(not sklearn_installed, reason="sklearn not installed")
    def test_bss_supported_return_false(self):
        assert self.s.blind_source_separation(return_info=False) is None



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import dask.array as da
import numpy as np
import pytest

from hyperspy.decorators import lazifyTestClass
from hyperspy.signals import BaseSignal, Signal2D


@lazifyTestClass
class TestTranspose:
    def setup_method(self, method):
        self.s = BaseSignal(np.random.rand(1, 2, 3, 4, 5, 6))
        for ax, name in zip(self.s.axes_manager._axes, "abcdef"):
            ax.name = name
        # just to make sure in case default changes
        self.s.axes_manager.set_signal_dimension(6)
        self.s.estimate_poissonian_noise_variance()

    def test_signal_int_transpose(self):
        t = self.s.transpose(signal_axes=2)
        var = t.metadata.Signal.Noise_properties.variance
        assert t.axes_manager.signal_shape == (6, 5)
        assert var.axes_manager.signal_shape == (6, 5)
        assert [ax.name for ax in t.axes_manager.signal_axes] == ["f", "e"]
        assert isinstance(t, Signal2D)
        assert isinstance(t.metadata.Signal.Noise_properties.variance, Signal2D)

    def test_signal_iterable_int_transpose(self):
        t = self.s.transpose(signal_axes=[0, 5, 4])
        var = t.metadata.Signal.Noise_properties.variance
        assert t.axes_manager.signal_shape == (6, 1, 2)
        assert var.axes_manager.signal_shape == (6, 1, 2)
        assert [ax.name for ax in t.axes_manager.signal_axes] == ["f", "a", "b"]

    def test_signal_iterable_names_transpose(self):
        t = self.s.transpose(signal_axes=["f", "a", "b"])
        var = t.metadata.Signal.Noise_properties.variance
        assert t.axes_manager.signal_shape == (6, 1, 2)
        assert var.axes_manager.signal_shape == (6, 1, 2)
        assert [ax.name for ax in t.axes_manager.signal_axes] == ["f", "a", "b"]

    def test_signal_iterable_axes_transpose(self):
        t = self.s.transpose(signal_axes=self.s.axes_manager.signal_axes[:2])
        var = t.metadata.Signal.Noise_properties.variance
        assert t.axes_manager.signal_shape == (6, 5)
        assert var.axes_manager.signal_shape == (6, 5)
        assert [ax.name for ax in t.axes_manager.signal_axes] == ["f", "e"]

    def test_signal_one_name(self):
        with pytest.raises(ValueError):
            self.s.transpose(signal_axes="a")

    def test_too_many_signal_axes(self):
        with pytest.raises(ValueError):
            self.s.transpose(signal_axes=10)

    def test_navigation_int_transpose(self):
        t = self.s.transpose(navigation_axes=2)
        var = t.metadata.Signal.Noise_properties.variance
        assert t.axes_manager.navigation_shape == (2, 1)
        assert var.axes_manager.navigation_shape == (2, 1)
        assert [ax.name for ax in t.axes_manager.navigation_axes] == ["b", "a"]

    def test_navigation_iterable_int_transpose(self):
        t = self.s.transpose(navigation_axes=[0, 5, 4])
        var = t.metadata.Signal.Noise_properties.variance
        assert t.axes_manager.navigation_shape == (6, 1, 2)
        assert var.axes_manager.navigation_shape == (6, 1, 2)
        assert [ax.name for ax in t.axes_manager.navigation_axes] == ["f", "a", "b"]

    def test_navigation_iterable_names_transpose(self):
        t = self.s.transpose(navigation_axes=["f", "a", "b"])
        var = t.metadata.Signal.Noise_properties.variance
        assert var.axes_manager.navigation_shape == (6, 1, 2)
        assert t.axes_manager.navigation_shape == (6, 1, 2)
        assert [ax.name for ax in t.axes_manager.navigation_axes] == ["f", "a", "b"]

    def test_navigation_iterable_axes_transpose(self):
        t = self.s.transpose(navigation_axes=self.s.axes_manager.signal_axes[:2])
        var = t.metadata.Signal.Noise_properties.variance
        assert t.axes_manager.navigation_shape == (6, 5)
        assert var.axes_manager.navigation_shape == (6, 5)
        assert [ax.name for ax in t.axes_manager.navigation_axes] == ["f", "e"]

    def test_navigation_one_name(self):
        with pytest.raises(ValueError):
            self.s.transpose(navigation_axes="a")

    def test_too_many_navigation_axes(self):
        with pytest.raises(ValueError):
            self.s.transpose(navigation_axes=10)

    def test_transpose_shortcut(self):
        s = self.s.transpose(signal_axes=2)
        t = s.T
        assert t.axes_manager.navigation_shape == (6, 5)
        assert [ax.name for ax in t.axes_manager.navigation_axes] == ["f", "e"]

    def test_optimize(self):
        if self.s._lazy:
            pytest.skip(
                "LazySignal optimization is tested in test_lazy_tranpose_rechunk"
            )
        t = self.s.transpose(signal_axes=["f", "a", "b"], optimize=False)
        assert t.data.base is self.s.data

        t = self.s.transpose(signal_axes=["f", "a", "b"], optimize=True)
        assert t.data.base is not self.s.data


def test_lazy_transpose_rechunks():
    ar = da.ones((50, 50, 256, 256), chunks=(5, 5, 256, 256))
    s = Signal2D(ar).as_lazy()
    s1 = s.T  # By default it does not rechunk
    cks = s.data.chunks
    assert s1.data.chunks == (cks[2], cks[3], cks[0], cks[1])
    s2 = s.transpose(optimize=True)
    assert s2.data.chunks != s1.data.chunks



# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
import numpy as np
import pytest
import scipy.ndimage
import traits.api as t

import hyperspy.api as hs
from hyperspy.drawing.utils import make_cmap, plot_RGB_map
from hyperspy.tests.drawing.test_plot_signal import _TestPlot

scalebar_color = 'blue'
default_tol = 2.0
baseline_dir = 'plot_signal2d'
style_pytest_mpl = 'default'


def _generate_image_stack_signal():
    image = hs.signals.Signal2D(np.random.random((2, 3, 512, 512)))
    for i in range(2):
        for j in range(3):
            image.data[i, j, :] = scipy.misc.ascent() * (i + 0.5 + j)
    axes = image.axes_manager
    axes[2].name = "x"
    axes[3].name = "y"
    axes[2].units = "nm"
    axes[3].units = "nm"

    return image


def _set_navigation_axes(axes_manager, name=t.Undefined, units=t.Undefined,
                         scale=1.0, offset=0.0):
    for nav_axis in axes_manager.navigation_axes:
        nav_axis.units = units
        nav_axis.scale = scale
        nav_axis.offset = offset
    return axes_manager


def _set_signal_axes(axes_manager, name=t.Undefined, units=t.Undefined,
                     scale=1.0, offset=0.0):
    for sig_axis in axes_manager.signal_axes:
        sig_axis.name = name
        sig_axis.units = units
        sig_axis.scale = scale
        sig_axis.offset = offset
    return axes_manager


@pytest.mark.parametrize("normalization", ['single', 'global'])
@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_rgb_image(normalization):
    w = 20
    data = np.arange(1, w * w + 1).reshape(w, w)
    ch1 = hs.signals.Signal2D(data)
    ch1.axes_manager = _set_signal_axes(ch1.axes_manager)
    ch2 = hs.signals.Signal2D(data.T * 2)
    ch2.axes_manager = _set_signal_axes(ch2.axes_manager)
    plot_RGB_map([ch1, ch2], normalization=normalization)
    return plt.gcf()


def _generate_parameter():
    parameters = []
    for scalebar in [True, False]:
        for colorbar in [True, False]:
            for axes_ticks in [True, False]:
                for centre_colormap in [True, False]:
                    for min_aspect in [0.2, 0.7]:
                        parameters.append([scalebar, colorbar, axes_ticks,
                                           centre_colormap, min_aspect])
    return parameters


@pytest.mark.parametrize(("scalebar", "colorbar", "axes_ticks",
                          "centre_colormap", "min_aspect"),
                         _generate_parameter())
@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot(scalebar, colorbar, axes_ticks, centre_colormap, min_aspect):
    test_plot = _TestPlot(ndim=0, sdim=2)
    test_plot.signal.plot(scalebar=scalebar,
                          colorbar=colorbar,
                          axes_ticks=axes_ticks,
                          centre_colormap=centre_colormap,
                          min_aspect=min_aspect)
    return test_plot.signal._plot.signal_plot.figure


def _generate_parameter_plot_images():
    # There are 9 images in total
    vmin, vmax = [None] * 9, [None] * 9
    vmin[1], vmax[2] = 30, 200
    return vmin, vmax


@pytest.mark.parametrize("percentile", [(None, None), ("1th", "99th")])
@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_log_scale(percentile):
    test_plot = _TestPlot(ndim=0, sdim=2)
    test_plot.signal += 1  # need to avoid zeros in log
    test_plot.signal.plot(norm='log', vmin=percentile[0], vmax=percentile[1])
    return test_plot.signal._plot.signal_plot.figure


@pytest.mark.parametrize("fft_shift", [True, False])
@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_FFT(fft_shift):
    s = hs.datasets.example_signals.object_hologram()
    s2 = s.isig[:128, :128].fft()
    s2.plot(fft_shift=fft_shift, axes_ticks=True, power_spectrum=True)
    return s2._plot.signal_plot.figure


@pytest.mark.parametrize(("vmin", "vmax"), (_generate_parameter_plot_images(),
                                            (None, None)))
@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_multiple_images_list(vmin, vmax):
    # load red channel of raccoon as an image
    image0 = hs.signals.Signal2D(scipy.misc.face()[:, :, 0])
    image0.metadata.General.title = 'Rocky Raccoon - R'
    axes0 = image0.axes_manager
    axes0[0].name = "x"
    axes0[1].name = "y"
    axes0[0].units = "mm"
    axes0[1].units = "mm"

    # load ascent into 2x3 hyperimage
    image1 = _generate_image_stack_signal()

    # load green channel of raccoon as an image
    image2 = hs.signals.Signal2D(scipy.misc.face()[:, :, 1])
    image2.metadata.General.title = 'Rocky Raccoon - G'
    axes2 = image2.axes_manager
    axes2[0].name = "x"
    axes2[1].name = "y"
    axes2[0].units = "mm"
    axes2[1].units = "mm"

    # load rgb imimagesage
    rgb = hs.signals.Signal1D(scipy.misc.face())
    rgb.change_dtype("rgb8")
    rgb.metadata.General.title = 'RGB'
    axesRGB = rgb.axes_manager
    axesRGB[0].name = "x"
    axesRGB[1].name = "y"
    axesRGB[0].units = "nm"
    axesRGB[1].units = "nm"

    hs.plot.plot_images([image0, image1, image2, rgb], tight_layout=True,
                        labelwrap=20, vmin=vmin, vmax=vmax)
    return plt.gcf()

@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_rgb_image():
    # load rgb imimagesage
    rgb = hs.signals.Signal1D(scipy.misc.face())
    rgb.change_dtype("rgb8")
    rgb.metadata.General.title = 'RGB'
    axesRGB = rgb.axes_manager
    axesRGB[0].name = "x"
    axesRGB[1].name = "y"
    axesRGB[0].units = "cm"
    axesRGB[1].units = "cm"
    rgb.plot()
    return plt.gcf()

class _TestIteratedSignal:

    def __init__(self):
        s = hs.signals.Signal2D([scipy.misc.ascent()] * 6)
        angles = hs.signals.BaseSignal(range(00, 60, 10))
        s.map(scipy.ndimage.rotate, angle=angles.T, reshape=False)
        # prevent values outside of integer range
        s.data = np.clip(s.data, 0, 255)
        title = 'Ascent'

        s.axes_manager = self._set_signal_axes(s.axes_manager,
                                               name='spatial',
                                               units='nm', scale=1,
                                               offset=0.0)
        s.axes_manager = self._set_navigation_axes(s.axes_manager,
                                                   name='index',
                                                   units='images',
                                                   scale=1, offset=0)
        s.metadata.General.title = title

        self.signal = s

    def _set_navigation_axes(self, axes_manager, name=t.Undefined,
                             units=t.Undefined, scale=1.0, offset=0.0):
        for nav_axis in axes_manager.navigation_axes:
            nav_axis.units = units
            nav_axis.scale = scale
            nav_axis.offset = offset
        return axes_manager

    def _set_signal_axes(self, axes_manager, name=t.Undefined,
                         units=t.Undefined, scale=1.0, offset=0.0):
        for sig_axis in axes_manager.signal_axes:
            sig_axis.name = name
            sig_axis.units = units
            sig_axis.scale = scale
            sig_axis.offset = offset
        return axes_manager


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_default():
    test_im_plot = _TestIteratedSignal()
    hs.plot.plot_images(test_im_plot.signal)
    return plt.gcf()


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_cmap_list():
    test_im_plot = _TestIteratedSignal()
    hs.plot.plot_images(test_im_plot.signal,
                        axes_decor='off',
                        cmap=['viridis', 'gray'])
    return plt.gcf()


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_cmap_list_w_diverging():
    test_im_plot = _TestIteratedSignal()
    hs.plot.plot_images(test_im_plot.signal,
                        axes_decor='off',
                        cmap=['viridis', 'gray', 'RdBu_r'])
    return plt.gcf()


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_cmap_mpl_colors():
    test_im_plot = _TestIteratedSignal()
    hs.plot.plot_images(test_im_plot.signal,
                        axes_decor='off',
                        cmap='mpl_colors')
    return plt.gcf()


def test_plot_images_cmap_mpl_colors_w_single_cbar():
    # This should give an error, so test for that
    test_im_plot = _TestIteratedSignal()
    with pytest.raises(ValueError):
        hs.plot.plot_images(test_im_plot.signal,
                            axes_decor='off',
                            cmap='mpl_colors',
                            colorbar='single')


def test_plot_images_bogus_cmap():
    # This should give an error, so test for that
    test_im_plot = _TestIteratedSignal()
    with pytest.raises(ValueError) as val_error:
        hs.plot.plot_images(test_im_plot.signal,
                            axes_decor='off',
                            cmap=3.14159265359,
                            colorbar=None)
    assert str(val_error.value) == 'The provided cmap value was not ' \
                                   'understood. Please check input values.'


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_cmap_one_string():
    test_im_plot = _TestIteratedSignal()
    hs.plot.plot_images(test_im_plot.signal,
                        axes_decor='off',
                        cmap='RdBu_r',
                        colorbar='single')
    return plt.gcf()


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_cmap_make_cmap_bittrue():
    test_im_plot = _TestIteratedSignal()
    hs.plot.plot_images(test_im_plot.signal,
                        axes_decor='off',
                        cmap=make_cmap([(255, 255, 255),
                                        '#F5B0CB',
                                        (220, 106, 207),
                                        '#745C97',
                                        (57, 55, 91)],
                                       bit=True,
                                       name='test_cmap',
                                       register=True))
    return plt.gcf()


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_cmap_make_cmap_bitfalse():
    test_im_plot = _TestIteratedSignal()
    hs.plot.plot_images(test_im_plot.signal,
                        axes_decor='off',
                        cmap=make_cmap([(1, 1, 1),
                                        '#F5B0CB',
                                        (0.86, 0.42, 0.81),
                                        '#745C97',
                                        (0.22, 0.22, 0.36)],
                                       bit=False,
                                       name='test_cmap',
                                       register=True))
    return plt.gcf()


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_cmap_multi_signal():
    test_plot1 = _TestIteratedSignal()

    test_plot2 = _TestIteratedSignal()
    test_plot2.signal *= 2  # change scale of second signal
    test_plot2.signal = test_plot2.signal.inav[::-1]
    test_plot2.signal.metadata.General.title = 'Descent'

    hs.plot.plot_images([test_plot1.signal, test_plot2.signal],
                        axes_decor='off',
                        per_row=4,
                        cmap='mpl_colors')
    return plt.gcf()


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_cmap_multi_w_rgb():
    test_plot1 = _TestIteratedSignal()

    test_plot2 = _TestIteratedSignal()
    test_plot2.signal *= 2  # change scale of second signal
    test_plot2.signal.metadata.General.title = 'Ascent-2'

    rgb_sig = hs.signals.Signal1D(scipy.misc.face())
    rgb_sig.change_dtype('rgb8')
    rgb_sig.metadata.General.title = 'Racoon!'

    hs.plot.plot_images([test_plot1.signal, test_plot2.signal, rgb_sig],
                        axes_decor='off',
                        per_row=4,
                        cmap='mpl_colors')
    return plt.gcf()


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_single_image():
    image0 = hs.signals.Signal2D(np.arange(100).reshape(10, 10))
    image0.isig[5, 5] = 200
    image0.metadata.General.title = 'This is the title from the metadata'
    hs.plot.plot_images(image0, vmin="0.05th", vmax="99.95th")
    return plt.gcf()


@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_single_image_stack():
    image0 = hs.signals.Signal2D(np.arange(200).reshape(2, 10, 10))
    image0.isig[5, 5] = 200
    image0.metadata.General.title = 'This is the title from the metadata'
    hs.plot.plot_images(image0, vmin="0.05th", vmax="99.95th")
    return plt.gcf()


def test_plot_images_multi_signal_w_axes_replot():
    imdata = np.random.rand(3, 5, 5)
    imgs = hs.signals.Signal2D(imdata)
    img_list = [imgs, imgs.inav[:2], imgs.inav[0]]
    subplots = hs.plot.plot_images(img_list, axes_decor=None)
    f = plt.gcf()
    f.canvas.draw()
    f.canvas.flush_events()

    tests = []
    for axi in subplots:
        imi = axi.images[0].get_array()
        x, y = axi.transData.transform((2, 2))
        # Calling base class method because of backends
        plt.matplotlib.backends.backend_agg.FigureCanvasBase.button_press_event(
            f.canvas, x, y, 'left', True)
        fn = plt.gcf()
        tests.append(
            np.allclose(imi, fn.axes[0].images[0].get_array().data))
        plt.close(fn)
    assert np.alltrue(tests)
    return f


@pytest.mark.parametrize("percentile", [("2.5th", "97.5th"),
                                        [["0th", "10th", "20th"], ["100th", "90th", "80th"]],
                                        [["5th", "10th"], ["95th", "90th"]],
                                        [["5th", None, "10th"], ["95th", None, "90th"]],
                                        ])
@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_vmin_vmax_percentile(percentile):
    image0 = hs.signals.Signal2D(np.arange(100).reshape(10, 10))
    image0.isig[5, 5] = 200
    image0.metadata.General.title = 'This is the title from the metadata'
    ax = hs.plot.plot_images([image0, image0, image0],
                             vmin=percentile[0],
                             vmax=percentile[1],
                             axes_decor='off')
    return ax[0].figure


@pytest.mark.parametrize("vmin_vmax", [(50, 150),
                                       ([0, 10], [120, None])])
@pytest.mark.parametrize("colorbar", ['single', 'multi', None])
@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_images_colorbar(colorbar, vmin_vmax):
    print("vmin_vmax:", vmin_vmax)
    image0 = hs.signals.Signal2D(np.arange(100).reshape(10, 10))
    image0.isig[5, 5] = 200
    image0.metadata.General.title = 'This is the title from the metadata'
    ax = hs.plot.plot_images([image0, image0],
                             colorbar=colorbar,
                             vmin=vmin_vmax[0],
                             vmax=vmin_vmax[1],
                             axes_decor='ticks')
    return ax[0].figure


def test_plot_images_signal1D():
    image0 = hs.signals.Signal1D(np.arange(100).reshape(10, 10))
    with pytest.raises(ValueError):
        hs.plot.plot_images([image0, image0])


def test_plot_images_not_signal():
    data = np.arange(100).reshape(10, 10)
    with pytest.raises(ValueError):
        hs.plot.plot_images([data, data])

    with pytest.raises(ValueError):
        hs.plot.plot_images(data)

    with pytest.raises(ValueError):
        hs.plot.plot_images('not a list of signal')


def test_plot_images_tranpose():
    a = hs.signals.BaseSignal(np.arange(100).reshape(10, 10))
    b = hs.signals.BaseSignal(np.arange(100).reshape(10, 10)).T

    hs.plot.plot_images([a, b.T])
    hs.plot.plot_images([a, b])


# Ignore numpy warning about clipping np.nan values
@pytest.mark.filterwarnings("ignore:Passing `np.nan` to mean no clipping in np.clip")
def test_plot_with_non_finite_value():
    s = hs.signals.Signal2D(np.array([[np.nan, 2.0] for v in range(2)]))
    s.plot()
    s.axes_manager.events.indices_changed.trigger(s.axes_manager)

    s = hs.signals.Signal2D(np.array([[np.nan, np.nan] for v in range(2)]))
    s.plot()
    s.axes_manager.events.indices_changed.trigger(s.axes_manager)

    s = hs.signals.Signal2D(np.array([[-np.inf, np.nan] for v in range(2)]))
    s.plot()
    s.axes_manager.events.indices_changed.trigger(s.axes_manager)

    s = hs.signals.Signal2D(np.array([[np.inf, np.nan] for v in range(2)]))
    s.plot()
    s.axes_manager.events.indices_changed.trigger(s.axes_manager)


@pytest.mark.parametrize("cmap", ['gray', None])
@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_log_negative_value(cmap):
    s = hs.signals.Signal2D(np.arange(10*10).reshape(10, 10))
    s -= 5*10
    if cmap:
        s.plot(norm='log', cmap=cmap)
    else:
        s.plot(norm='log')
    return plt.gcf()


@pytest.mark.parametrize("cmap", ['gray', None, 'preference'])
@pytest.mark.mpl_image_compare(
    baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl)
def test_plot_navigator_colormap(cmap):
    if cmap == 'preference':
        hs.preferences.Plot.cmap_navigator = 'hot'
        cmap = None
    s = hs.signals.Signal1D(np.arange(10*10*10).reshape(10, 10, 10))
    s.plot(navigator_kwds={'cmap':cmap})
    return s._plot.navigator_plot.figure


@pytest.mark.parametrize("autoscale", ['', 'xy', 'xv', 'xyv', 'v'])
@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir,
                               tolerance=default_tol, style=style_pytest_mpl)
def test_plot_autoscale(autoscale):
    s = hs.signals.Signal2D(np.arange(100).reshape(10, 10))
    s.plot(autoscale=autoscale, axes_ticks=True)
    imf = s._plot.signal_plot
    ax = imf.ax
    extend = [5.0, 10.0, 3., 10.0]
    ax.images[0].set_extent(extend)
    ax.set_xlim(5.0, 10.0)
    ax.set_ylim(3., 10.0)

    ax.images[0].norm.vmin = imf._vmin = 10
    ax.images[0].norm.vmax = imf._vmax = 50

    s.axes_manager.events.indices_changed.trigger(s.axes_manager)
    # Because we are hacking the vmin, vmax with matplotlib, we need to update
    # colorbar too
    imf._colorbar.draw_all()

    return s._plot.signal_plot.figure


@pytest.mark.parametrize("autoscale", ['', 'v'])
def test_plot_autoscale_data_changed(autoscale):
    s = hs.signals.Signal2D(np.arange(100).reshape(10, 10))
    s.plot(autoscale=autoscale, axes_ticks=True)
    imf = s._plot.signal_plot
    _vmin = imf._vmin
    _vmax = imf._vmax

    s.data = s.data / 2
    s.events.data_changed.trigger(s)

    if 'v' in autoscale:
        np.testing.assert_allclose(imf._vmin, s.data.min())
        np.testing.assert_allclose(imf._vmax, s.data.max())
    else:
        np.testing.assert_allclose(imf._vmin, _vmin)
        np.testing.assert_allclose(imf._vmax, _vmax)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import matplotlib.pyplot as plt
import logging

from hyperspy.drawing.widgets import Widget2DBase, ResizersMixin


_logger = logging.getLogger(__name__)
# Track if we have already warned when the widget is out of range
already_warn_out_of_range = False


class SquareWidget(Widget2DBase):

    """SquareWidget is a symmetric, Rectangle-patch based widget, which can be
    dragged, and resized by keystrokes/code. As the widget is normally only
    meant to indicate position, the sizing is deemed purely visual, but there
    is nothing that forces this use. However, it should be noted that the outer
    bounds only correspond to pure indices for odd sizes.
    """

    def __init__(self, axes_manager, **kwargs):
        super(SquareWidget, self).__init__(axes_manager, **kwargs)

    def _set_patch(self):
        """Sets the patch to a matplotlib Rectangle with the correct geometry.
        The geometry is defined by _get_patch_xy, and get_size_in_axes.
        """
        xy = self._get_patch_xy()
        xs, ys = self.size
        self.patch = [plt.Rectangle(
            xy, xs, ys,
            animated=self.blit,
            fill=False,
            lw=self.border_thickness,
            ec=self.color,
            alpha=self.alpha,
            picker=True,)]
        super(SquareWidget, self)._set_patch()

    def _onmousemove(self, event):
        """on mouse motion move the patch if picked"""
        if self.picked is True and event.inaxes:
            self.position = (event.xdata, event.ydata)


class RectangleWidget(SquareWidget, ResizersMixin):

    """RectangleWidget is a asymmetric, Rectangle-patch based widget, which can
    be dragged and resized by mouse/keys. For resizing by mouse, it adds a
    small Rectangle patch on the outer border of the main patch, to serve as
    resize handles. This feature can be enabled/disabled by the 'resizers'
    property, and the size/color of the handles are set by
    'resize_color'/'resize_pixel_size'.

    For optimized changes of geometry, the class implements two methods
    'set_bounds' and 'set_ibounds', to set the geomtry of the rectangle by
    value and index space coordinates, respectivly. It also adds the 'width'
    and 'height' properties for verbosity.

    For keyboard resizing, 'x'/'c' and 'y'/'u' will increase/decrease the size
    of the rectangle along the first and the second axis, respectively.

    Implements the internal method _validate_geometry to make sure the patch
    will always stay within bounds.
    """

    # --------- External interface ---------

    def _parse_bounds_args(self, args, kwargs):
        """Internal utility function to parse args/kwargs passed to set_bounds
        and set_ibounds.
        """
        if len(args) == 1:
            return args[0]
        elif len(args) == 4:
            return args
        elif len(kwargs) == 1 and 'bounds' in kwargs:
            return kwargs.values()[0]
        else:
            x = kwargs.pop('x', kwargs.pop('left', self._pos[0]))
            y = kwargs.pop('y', kwargs.pop('top', self._pos[1]))
            if 'right' in kwargs:
                w = kwargs.pop('right') - x
            else:
                w = kwargs.pop('w', kwargs.pop('width', self._size[0]))
            if 'bottom' in kwargs:
                h = kwargs.pop('bottom') - y
            else:
                h = kwargs.pop('h', kwargs.pop('height', self._size[1]))
            return x, y, w, h

    def set_ibounds(self, *args, **kwargs):
        """
        Set bounds by indices. Bounds can either be specified in order left,
        bottom, width, height; or by keywords:

        * 'bounds': tuple (left, top, width, height)

        OR

        * 'x'/'left'
        * 'y'/'top'
        * 'w'/'width', alternatively 'right'
        * 'h'/'height', alternatively 'bottom'

        If specifying with keywords, any unspecified dimensions will be kept
        constant (note: width/height will be kept, not right/bottom).
        """

        ix, iy, iw, ih = self._parse_bounds_args(args, kwargs)

        x = self.axes[0].index2value(ix)
        y = self.axes[1].index2value(iy)
        w = self._i2v(self.axes[0], ix + iw) - x
        h = self._i2v(self.axes[1], iy + ih) - y

        old_position, old_size = self.position, self.size
        self._pos = np.array([x, y])
        self._size = np.array([w, h])
        self._apply_changes(old_size=old_size, old_position=old_position)

    def set_bounds(self, *args, **kwargs):
        """
        Set bounds by values. Bounds can either be specified in order left,
        bottom, width, height; or by keywords:

        * 'bounds': tuple (left, top, width, height)

        OR

        * 'x'/'left'
        * 'y'/'top'
        * 'w'/'width', alternatively 'right' (x+w)
        * 'h'/'height', alternatively 'bottom' (y+h)

        If specifying with keywords, any unspecified dimensions will be kept
        constant (note: width/height will be kept, not right/bottom).
        """
        global already_warn_out_of_range
        x, y, w, h = self._parse_bounds_args(args, kwargs)

        def warn(obj, parameter, value):
            global already_warn_out_of_range
            if not already_warn_out_of_range:
                _logger.info('{}: {} is out of range. It is therefore set '
                             'to the value of {}'.format(obj, parameter, value))
                already_warn_out_of_range = True

        scale = [axis.scale for axis in self.axes]
        l0, h0 = self.axes[0].low_value, self.axes[0].high_value
        l1, h1 = self.axes[1].low_value, self.axes[1].high_value

        in_range = 0
        if x < l0:
            x = l0
            warn(self, '`x`', x)
        elif h0 <= x:
            x = h0 - scale[0]
            warn(self, '`x`', x)
        else:
            in_range += 1
        if y < l1:
            y = l1
            warn(self, '`y`', y)
        elif h1 <= y:
            warn(self, '`y`', y)
            y = h1 - scale[1]
        else:
            in_range += 1
        if w < scale[0]:
            w = scale[0]
            warn(self, '`width` or `right`', w)
        elif not (l0 + scale[0] <= x + w <= h0 + scale[0]):
            if self.size[0] != w:  # resize
                w = h0 + scale[0] - self.position[0]
                warn(self, '`width` or `right`', w)
            if self.position[0] != x:  # moved
                x = h0 + scale[0] - self.size[0]
                warn(self, '`x`', x)
        else:
            in_range += 1
        if h < scale[1]:
            h = scale[1]
            warn(self, '`height` or `bottom`', h)
        elif not (l1 + scale[1] <= y + h <= h1 + scale[1]):
            if self.size[1] != h:  # resize
                h = h1 + scale[1] - self.position[1]
                warn(self, '`height` or `bottom`', h)
            if self.position[1] != y:  # moved
                y = h1 + scale[1] - self.size[1]
                warn(self, '`y`', y)
        else:
            in_range += 1

        # if we are in range again, reset `already_warn_out_of_range` to False
        if in_range == 4 and already_warn_out_of_range:
            _logger.info('{} back in range.'.format(self.__class__.__name__))
            already_warn_out_of_range = False

        old_position, old_size = self.position, self.size
        self._pos = np.array([x, y])
        self._size = np.array([w, h])
        self._apply_changes(old_size=old_size, old_position=old_position)

    def _validate_pos(self, value):
        """Constrict the position within bounds.
        """
        value = (min(value[0], self.axes[0].high_value - self._size[0] +
                     self.axes[0].scale),
                 min(value[1], self.axes[1].high_value - self._size[1] +
                     self.axes[1].scale))
        return super(RectangleWidget, self)._validate_pos(value)

    @property
    def width(self):
        return self.get_size_in_indices()[0]

    @width.setter
    def width(self, value):
        if value == self.width:
            return
        ix = self.indices[0] + value
        il0, ih0 = self.axes[0].low_index, self.axes[0].high_index
        if value <= 0 or not (il0 < ix <= ih0):
            raise ValueError('`width` value is not in range. The '
                             '`width` is {} and should be in range '
                             '{}-{}.'.format(ix, il0 + 1, ih0))
        self._set_a_size(0, value)

    @property
    def height(self):
        return self.get_size_in_indices()[1]

    @height.setter
    def height(self, value):
        if value == self.height:
            return
        iy = self.indices[1] + value
        il1, ih1 = self.axes[1].low_index, self.axes[1].high_index
        if value <= 0 or not (il1 < iy <= ih1):
            raise ValueError('`height` value is not in range. The '
                             '`height` is {} and should be in range '
                             '{}-{}.'.format(iy, il1 + 1, ih1))
        self._set_a_size(1, value)

    # --------- Internal functions ---------

    # --- Internals that trigger events ---

    def _set_size(self, value):
        value = np.minimum(value, [ax.size * ax.scale for ax in self.axes])
        value = np.maximum(value, [ax.scale for ax in self.axes])
        if np.any(self._size != value):
            old = self._size
            self._size = value
            self._validate_geometry()
            if np.any(self._size != old):
                self._size_changed()

    def _set_a_size(self, idx, value):
        if self._size[idx] == value or value <= 0:
            return
        # If we are pushed "past" an edge, size towards it
        if self._navigating and self.axes[idx].value > self._pos[idx]:
            if value < self._size[idx]:
                self._pos[idx] += self._size[idx] - value

        self._size[idx] = value
        self._validate_geometry()
        self._size_changed()

    def _increase_xsize(self):
        self._set_a_size(0, self._size[0] +
                         self.axes[0].scale * self.size_step)

    def _decrease_xsize(self):
        new_s = self._size[0] - self.axes[0].scale * self.size_step
        new_s = max(new_s, self.axes[0].scale)
        self._set_a_size(0, new_s)

    def _increase_ysize(self):
        self._set_a_size(1, self._size[1] +
                         self.axes[1].scale * self.size_step)

    def _decrease_ysize(self):
        new_s = self._size[1] - self.axes[1].scale * self.size_step
        new_s = max(new_s, self.axes[1].scale)
        self._set_a_size(1, new_s)

    def on_key_press(self, event):
        if self.selected:
            if event.key == "x":
                self._increase_xsize()
            elif event.key == "c":
                self._decrease_xsize()
            elif event.key == "y":
                self._increase_ysize()
            elif event.key == "u":
                self._decrease_ysize()
            else:
                super(RectangleWidget, self).on_key_press(event)

    # --- End internals that trigger events ---

    def _get_patch_xy(self):
        """Get xy value for Rectangle with position being top left. This value
        deviates from the 'position', as 'position' correspond to the center
        value of the pixel. Here, xy corresponds to the top left of the pixel.
        """
        offset = [a.scale for a in self.axes]
        return self._pos - 0.5 * np.array(offset)

    def _update_patch_position(self):
        # Override to include resizer positioning
        if self.is_on() and self.patch:
            self.patch[0].set_xy(self._get_patch_xy())
            self._update_resizers()
            self.draw_patch()

    def _update_patch_geometry(self):
        # Override to include resizer positioning
        if self.is_on() and self.patch:
            self.patch[0].set_bounds(*self._get_patch_bounds())
            self._update_resizers()
            self.draw_patch()

    def _validate_geometry(self, x1=None, y1=None):
        """Make sure the entire patch always stays within bounds. First the
        position (either from position property or from x1/y1 arguments), is
        limited within the bounds. Then, if the bottom/right edges are out of
        bounds, the position is changed so that they will be at the limit.

        The modified geometry is stored, but no change checks are performed.
        Call _apply_changes after this in order to process any changes (the
        size might change if it is set larger than the bounds size).
        """
        xaxis = self.axes[0]
        yaxis = self.axes[1]

        # Make sure widget size is not larger than axes
        self._size[0] = min(self._size[0], xaxis.size * xaxis.scale)
        self._size[1] = min(self._size[1], yaxis.size * yaxis.scale)

        # Make sure x1/y1 is within bounds
        if x1 is None:
            x1 = self._pos[0]  # Get it if not supplied
        elif x1 < xaxis.low_value:
            x1 = xaxis.low_value
        elif x1 > xaxis.high_value:
            x1 = xaxis.high_value

        if y1 is None:
            y1 = self._pos[1]
        elif y1 < yaxis.low_value:
            y1 = yaxis.low_value
        elif y1 > yaxis.high_value:
            y1 = yaxis.high_value

        # Make sure x2/y2 is with upper bound.
        # If not, keep dims, and change x1/y1!
        x2 = x1 + self._size[0]
        y2 = y1 + self._size[1]
        if x2 > xaxis.high_value + xaxis.scale:
            x2 = xaxis.high_value + xaxis.scale
            x1 = x2 - self._size[0]
        if y2 > yaxis.high_value + yaxis.scale:
            y2 = yaxis.high_value + yaxis.scale
            y1 = y2 - self._size[1]

        self._pos = np.array([x1, y1])
        # Apply snaps if appropriate
        if self.snap_position:
            self._do_snap_position()
        if self.snap_size:
            self._do_snap_size()

    def _onmousemove(self, event):
        """on mouse motion draw the cursor if picked"""
        # Simple checks to make sure we are dragging our patch:
        if self.picked is True and event.inaxes:
            # Setup reused parameters
            xaxis = self.axes[0]
            yaxis = self.axes[1]
            # Mouse position
            x = event.xdata
            y = event.ydata
            p = self._get_patch_xy()
            # Old bounds
            bounds = [p[0], p[1],
                      p[0] + self._size[0],
                      p[1] + self._size[1]]

            # Store geometry for _apply_changes at end
            old_position, old_size = self.position, self.size

            if self.resizer_picked is False:
                # Simply dragging main patch. Offset mouse position by
                # pick_offset to get new position, then validate it.
                x -= self.pick_offset[0]
                y -= self.pick_offset[1]
                self._validate_geometry(x, y)
            else:
                posx = None     # New x pos. If None, the old pos will be used
                posy = None     # Same for y
                corner = self.resizer_picked
                # Adjust for resizer position:
                offset = self._get_resizer_offset()
                x += offset[0] * (0.5 - 1 * (corner % 2))
                y += offset[1] * (0.5 - 1 * (corner // 2))

                if corner % 2 == 0:         # Left side start
                    if x > bounds[2]:       # flipped to right
                        posx = bounds[2]    # New left is old right
                        # New size is mouse position - new left
                        self._size[0] = x - posx
                        self.resizer_picked += 1         # Switch pick to right
                    elif bounds[2] - x < xaxis.scale:   # Width too small
                        posx = bounds[2] - xaxis.scale      # So move pos left
                        self._size[0] = bounds[2] - posx    # Should be scale
                    else:                   # Moving left edge
                        posx = x            # Set left to mouse position
                        # Keep right still by changing size:
                        self._size[0] = bounds[2] - x
                else:                       # Right side start
                    if x < bounds[0]:       # Flipped to left
                        if bounds[0] - x < xaxis.scale:
                            posx = bounds[0] - xaxis.scale
                        else:
                            posx = x            # Set left to mouse
                        # Set size to old left - new left
                        self._size[0] = bounds[0] - posx
                        self.resizer_picked -= 1     # Switch pick to left
                    else:                   # Moving right edge
                        # Left should be left as it is, only size updates:
                        self._size[0] = x - bounds[0]  # mouse - old left
                if corner // 2 == 0:        # Top side start
                    if y > bounds[3]:       # flipped to botton
                        posy = bounds[3]    # New top is old bottom
                        # New size is mouse position - new top
                        self._size[1] = y - posy
                        self.resizer_picked += 2     # Switch pick to bottom
                    elif bounds[3] - y < yaxis.scale:  # Height too small
                        posy = bounds[3] - yaxis.scale      # So move pos up
                        self._size[1] = bounds[3] - posy    # Should be scale
                    else:                   # Moving top edge
                        posy = y           # Set top to mouse index
                        # Keep bottom still by changing size:
                        self._size[1] = bounds[3] - y  # old bottom - new top
                else:                       # Bottom side start
                    if y < bounds[1]:     # Flipped to top
                        if bounds[1] - y < yaxis.scale:
                            posy = bounds[1] - yaxis.scale
                        else:
                            posy = y           # Set top to mouse
                        # Set size to old top - new top
                        self._size[1] = bounds[1] - posy
                        self.resizer_picked -= 2     # Switch pick to top
                    else:                   # Moving bottom edge
                        self._size[1] = y - bounds[1]  # mouse - old top
                # Bound size to scale:
                if self._size[0] < xaxis.scale:
                    self._size[0] = xaxis.scale
                if self._size[1] < yaxis.scale:
                    self._size[1] = yaxis.scale
                if posx is not None:
                    posx += 0.5 * xaxis.scale
                if posy is not None:
                    posy += 0.5 * yaxis.scale
                # Validate the geometry
                self._validate_geometry(posx, posy)
            # Finally, apply any changes and trigger events/redraw:
            self._apply_changes(old_size=old_size, old_position=old_position)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import dask.array as da

from hyperspy._components.expression import Expression
from distutils.version import LooseVersion
import sympy

sqrt2pi = np.sqrt(2 * np.pi)


def _estimate_skewnormal_parameters(signal, x1, x2, only_current):
    axis = signal.axes_manager.signal_axes[0]
    i1, i2 = axis.value_range_to_indices(x1, x2)
    X = axis.axis[i1:i2]
    if only_current is True:
        data = signal()[i1:i2]
        X_shape = (len(X),)
        i = 0
        x0_shape = (1,)
    else:
        i = axis.index_in_array
        data_gi = [slice(None), ] * len(signal.data.shape)
        data_gi[axis.index_in_array] = slice(i1, i2)
        data = signal.data[tuple(data_gi)]
        X_shape = [1, ] * len(signal.data.shape)
        X_shape[axis.index_in_array] = data.shape[i]
        x0_shape = list(data.shape)
        x0_shape[i] = 1

    if isinstance(data, da.Array):
        _sum = da.sum
        _sqrt = da.sqrt
        _abs = abs
        _argmin = da.argmin

    else:
        _sum = np.sum
        _sqrt = np.sqrt
        _abs = np.abs
        _argmin = np.argmin

    a1 = np.sqrt(2 / np.pi)
    b1 = (4 / np.pi - 1) * a1
    m1 = _sum(X.reshape(X_shape) * data, i) / _sum(data, i)
    m2 = _abs(_sum((X.reshape(X_shape) - m1.reshape(x0_shape)) ** 2 * data, i)
              / _sum(data, i))
    m3 = _abs(_sum((X.reshape(X_shape) - m1.reshape(x0_shape)) ** 3 * data, i)
              / _sum(data, i))

    x0 = m1 - a1 * (m3 / b1) ** (1 / 3)
    scale = _sqrt(m2 + a1 ** 2 * (m3 / b1) ** (2 / 3))
    delta = _sqrt(1 / (a1**2 + m2 * (b1 / m3) ** (2 / 3)))
    shape = delta / _sqrt(1 - delta**2)

    iheight = _argmin(_abs(X.reshape(X_shape) - x0.reshape(x0_shape)), i)
    # height is the value of the function at x0, shich has to be computed
    # differently for dask array (lazy) and depending on the dimension
    if isinstance(data, da.Array):
        x0, iheight, scale, shape = da.compute(x0, iheight, scale, shape)
        if only_current is True or signal.axes_manager.navigation_dimension == 0:
            height = data.vindex[iheight].compute()
        elif signal.axes_manager.navigation_dimension == 1:
            height = data.vindex[np.arange(signal.axes_manager.navigation_size),
                                 iheight].compute()
        else:
            height = data.vindex[(*np.indices(signal.axes_manager.navigation_shape),
                                  iheight)].compute()
    else:
        if only_current is True or signal.axes_manager.navigation_dimension == 0:
            height = data[iheight]
        elif signal.axes_manager.navigation_dimension == 1:
            height = data[np.arange(signal.axes_manager.navigation_size),
                          iheight]
        else:
            height = data[(*np.indices(signal.axes_manager.navigation_shape),
                           iheight)]

    return x0, height, scale, shape


class SkewNormal(Expression):

    r"""Skew normal distribution component.

    |  Asymmetric peak shape based on a normal distribution.
    |  For definition see
       https://en.wikipedia.org/wiki/Skew_normal_distribution
    |  See also http://azzalini.stat.unipd.it/SN/
    |

    .. math::

        f(x) &= 2 A \phi(x) \Phi(x) \\
        \phi(x) &= \frac{1}{\sqrt{2\pi}}\mathrm{exp}{\left[
                   -\frac{t(x)^2}{2}\right]} \\
        \Phi(x) &= \frac{1}{2}\left[1 + \mathrm{erf}\left(\frac{
                   \alpha~t(x)}{\sqrt{2}}\right)\right] \\
        t(x) &= \frac{x-x_0}{\omega}


    ============== =============
    Variable        Parameter 
    ============== =============
    :math:`x_0`     x0 
    :math:`A`       A 
    :math:`\omega`  scale 
    :math:`\alpha`  shape 
    ============== =============


    Parameters
    -----------
    x0 : float
        Location of the peak position (not maximum, which is given by
        the `mode` property).
    A : float
        Height parameter of the peak.
    scale : float
        Width (sigma) parameter.
    shape: float
        Skewness (asymmetry) parameter. For shape=0, the normal
        distribution (Gaussian) is obtained. The distribution is
        right skewed (longer tail to the right) if shape>0 and is
        left skewed if shape<0.


    The properties `mean` (position), `variance`, `skewness` and `mode`
    (=position of maximum) are defined for convenience.
    """

    def __init__(self, x0=0., A=1., scale=1., shape=0.,
                 module=['numpy', 'scipy'], **kwargs):
        if LooseVersion(sympy.__version__) < LooseVersion("1.3"):
            raise ImportError("The `SkewNormal` component requires "
                              "SymPy >= 1.3")
        super(SkewNormal, self).__init__(
            expression="2 * A * normpdf * normcdf; normpdf = exp(- t ** 2 / 2) \
                / sqrt(2 * pi); normcdf = (1 + erf(shape * t / sqrt(2))) / 2; \
                t = (x - x0) / scale",
            name="SkewNormal",
            x0=x0,
            A=A,
            scale=scale,
            shape=shape,
            module=module,
            autodoc=False,
            **kwargs,
        )

        # Boundaries
        self.A.bmin = 0.

        self.scale.bmin = 0

        self.isbackground = False
        self.convolved = True

    def estimate_parameters(self, signal, x1, x2, only_current=False):
        """Estimate the skew normal distribution by calculating the momenta.

        Parameters
        ----------
        signal : Signal1D instance
        x1 : float
            Defines the left limit of the spectral range to use for the
            estimation.
        x2 : float
            Defines the right limit of the spectral range to use for the
            estimation.

        only_current : bool
            If False estimates the parameters for the full dataset.

        Returns
        -------
        bool

        Notes
        -----
        Adapted from Lin, Lee and Yen, Statistica Sinica 17, 909-927 (2007)
        https://www.jstor.org/stable/24307705

        Examples
        --------

        >>> g = hs.model.components1D.SkewNormal()
        >>> x = np.arange(-10, 10, 0.01)
        >>> data = np.zeros((32, 32, 2000))
        >>> data[:] = g.function(x).reshape((1, 1, 2000))
        >>> s = hs.signals.Signal1D(data)
        >>> s.axes_manager._axes[-1].offset = -10
        >>> s.axes_manager._axes[-1].scale = 0.01
        >>> g.estimate_parameters(s, -10, 10, False)
        """

        super(SkewNormal, self)._estimate_parameters(signal)
        axis = signal.axes_manager.signal_axes[0]
        x0, height, scale, shape = _estimate_skewnormal_parameters(signal, x1,
                                                                   x2, only_current)
        if only_current is True:
            self.x0.value = x0
            self.A.value = height * sqrt2pi
            self.scale.value = scale
            self.shape.value = shape
            if self.binned:
                self.A.value /= axis.scale
            return True
        else:
            if self.A.map is None:
                self._create_arrays()
            self.A.map['values'][:] = height * sqrt2pi

            if self.binned:
                self.A.map['values'] /= axis.scale
            self.A.map['is_set'][:] = True
            self.x0.map['values'][:] = x0
            self.x0.map['is_set'][:] = True
            self.scale.map['values'][:] = scale
            self.scale.map['is_set'][:] = True
            self.shape.map['values'][:] = shape
            self.shape.map['is_set'][:] = True
            self.fetch_stored_values()
            return True

    @property
    def mean(self):
        delta = self.shape.value / np.sqrt(1 + self.shape.value**2)
        return self.x0.value + self.scale.value * delta * np.sqrt(2 / np.pi)

    @property
    def variance(self):
        delta = self.shape.value / np.sqrt(1 + self.shape.value**2)
        return self.scale.value**2 * (1 - 2 * delta**2 / np.pi)

    @property
    def skewness(self):
        delta = self.shape.value / np.sqrt(1 + self.shape.value**2)
        return (4 - np.pi)/2 * (delta * np.sqrt(2/np.pi))**3 / (1 -
                                                                2 * delta**2 / np.pi)**(3/2)

    @property
    def mode(self):
        delta = self.shape.value / np.sqrt(1 + self.shape.value**2)
        muz = np.sqrt(2 / np.pi) * delta
        sigmaz = np.sqrt(1 - muz**2)
        if self.shape.value == 0:
            return self.x0.value
        else:
            m0 = muz - self.skewness * sigmaz / 2 - np.sign(self.shape.value) \
                / 2 * np.exp(- 2 * np.pi / np.abs(self.shape.value))
            return self.x0.value + self.scale.value * m0



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from hyperspy.signals import Signal2D
from hyperspy.utils import roi

BASELINE_DIR = 'plot_roi'
DEFAULT_TOL = 2.0
STYLE_PYTEST_MPL = 'default'


def _transpose_space(space, im):
    if space == "signal":
        im = im
        axes = im.axes_manager.signal_axes
        im.plot()
        figure = im._plot.signal_plot.figure
    else:
        im = im.T
        axes = im.axes_manager.navigation_axes
        im.plot()
        figure = im._plot.navigator_plot.figure
    return {
        "im": im,
        "figure": figure,
        "axes": axes,
    }


class TestPlotROI():

    def setup_method(self, method):
        # Create test image 100x100 pixels:
        self.im = Signal2D(np.arange(50000).reshape([10, 50, 100]))
        self.im.axes_manager[0].scale = 1e-1
        self.im.axes_manager[1].scale = 1e-2
        self.im.axes_manager[2].scale = 1e-3

    @pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR,
                                   tolerance=DEFAULT_TOL, style=STYLE_PYTEST_MPL)
    def test_plot_point1D_axis_0(self):
        self.im.plot()
        p = roi.Point1DROI(0.5)
        p.add_widget(signal=self.im, axes=[0, ], color="cyan")
        return self.im._plot.navigator_plot.figure

    @pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR,
                                   tolerance=DEFAULT_TOL, style=STYLE_PYTEST_MPL)
    def test_plot_point1D_axis_1(self):
        self.im.plot()
        p = roi.Point1DROI(0.05)
        p.add_widget(signal=self.im, axes=[1, ], color="cyan")
        return self.im._plot.signal_plot.figure

    @pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR,
                                   tolerance=DEFAULT_TOL, style=STYLE_PYTEST_MPL)
    def test_plot_point1D_axis_2(self):
        self.im.plot()
        p = roi.Point1DROI(0.005)
        p.add_widget(signal=self.im, axes=[2, ], color="cyan")
        return self.im._plot.signal_plot.figure

    @pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR,
                                   tolerance=DEFAULT_TOL, style=STYLE_PYTEST_MPL)
    def test_plot_spanroi_axis_0(self):
        self.im.plot()
        p = roi.SpanROI(0.5, 0.7)
        p.add_widget(signal=self.im, axes=[0, ], color="cyan")
        return self.im._plot.navigator_plot.figure

    @pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR,
                                   tolerance=DEFAULT_TOL, style=STYLE_PYTEST_MPL)
    def test_plot_spanroi_axis_1(self):
        self.im.plot()
        p = roi.SpanROI(0.05, 0.07)
        p.add_widget(signal=self.im, axes=[1, ], color="cyan")
        return self.im._plot.signal_plot.figure

    @pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR,
                                   tolerance=DEFAULT_TOL, style=STYLE_PYTEST_MPL)
    def test_plot_spanroi_axis_2(self):
        self.im.plot()
        p = roi.SpanROI(0.005, 0.007)
        p.add_widget(signal=self.im, axes=[2, ], color="cyan")
        return self.im._plot.signal_plot.figure

    @pytest.mark.parametrize("space", ("signal", "navigation"))
    @pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR,
                                   tolerance=DEFAULT_TOL, style=STYLE_PYTEST_MPL)
    def test_plot_point2D(self, space):
        objs = _transpose_space(im=self.im, space=space)
        p = roi.Point2DROI(0.05, 0.01)
        p.add_widget(signal=objs["im"], axes=objs["axes"], color="cyan")
        return objs["figure"]

    @pytest.mark.parametrize("space", ("signal", "navigation"))
    @pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR,
                                   tolerance=DEFAULT_TOL, style=STYLE_PYTEST_MPL)
    def test_plot_circle_roi(self, space):
        self.im.axes_manager[2].scale = 0.01
        objs = _transpose_space(im=self.im, space=space)
        p = roi.CircleROI(cx=0.1, cy=0.1, r=0.1)
        p.add_widget(signal=objs["im"], axes=objs["axes"], color="cyan")
        return objs["figure"]

    @pytest.mark.parametrize("space", ("signal", "navigation"))
    @pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR,
                                   tolerance=DEFAULT_TOL, style=STYLE_PYTEST_MPL)
    def test_plot_rectangular_roi(self, space):
        objs = _transpose_space(im=self.im, space=space)
        p = roi.RectangularROI(left=0.01, top=0.01, right=0.1, bottom=0.03)
        p.add_widget(signal=objs["im"], axes=objs["axes"], color="cyan")
        return objs["figure"]

    @pytest.mark.parametrize("space", ("signal", "navigation"))
    @pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR,
                                   tolerance=DEFAULT_TOL, style=STYLE_PYTEST_MPL)
    def test_plot_line2d_roi(self, space):
        objs = _transpose_space(im=self.im, space=space)
        p = roi.Line2DROI(x1=0.01, y1=0.01, x2=0.1, y2=0.03)
        p.add_widget(signal=objs["im"], axes=objs["axes"], color="cyan")
        return objs["figure"]


def test_error_message():
    im = Signal2D(np.arange(50000).reshape([10, 50, 100]))
    im.plot()
    im._plot.close()
    p = roi.Point1DROI(0.5)
    with pytest.raises(Exception, match='does not have an active plot.'):
        p.add_widget(signal=im, axes=[0, ], color="cyan")



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest
from hyperspy.misc.tv_denoise import tv_denoise
from scipy.misc import ascent


def test_tv_denoise_error():
    with pytest.raises(ValueError, match="may be denoised"):
        _ = tv_denoise(np.array([1, 2, 3]))


def test_2d_tv_denoise():
    rng = np.random.RandomState(123)
    data = ascent().astype(np.float)
    data_noisy = data + data.std() * rng.randn(*data.shape)
    data_clean = tv_denoise(data, weight=60)

    norm_noisy = np.linalg.norm(data - data_noisy) / np.linalg.norm(data)
    norm_clean = np.linalg.norm(data - data_clean) / np.linalg.norm(data)

    np.testing.assert_allclose(norm_noisy, 0.48604971)
    np.testing.assert_allclose(norm_clean, 0.10888393)


def test_3d_tv_denoise():
    rng = np.random.RandomState(123)
    x, y, z = np.ogrid[0:40, 0:40, 0:40]
    data = (x - 22) ** 2 + (y - 20) ** 2 + (z - 17) ** 2 < 8 ** 2
    data = 255 * data.astype(np.float)
    data_noisy = data + data.std() * rng.randn(*data.shape)
    data_clean = tv_denoise(data_noisy, weight=100)

    norm_noisy = np.linalg.norm(data - data_noisy) / np.linalg.norm(data)
    norm_clean = np.linalg.norm(data - data_clean) / np.linalg.norm(data)

    print(norm_clean)
    np.testing.assert_allclose(norm_noisy, 0.98151071)
    np.testing.assert_allclose(norm_clean, 0.12519535)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy import signals
from hyperspy.misc.utils import (
    is_hyperspy_signal,
    parse_quantity,
    slugify,
    strlist2enumeration,
)


def test_slugify():
    assert slugify("a") == "a"
    assert slugify("1a") == "1a"
    assert slugify("1") == "1"
    assert slugify("a a") == "a_a"

    assert slugify("a", valid_variable_name=True) == "a"
    assert slugify("1a", valid_variable_name=True) == "Number_1a"
    assert slugify("1", valid_variable_name=True) == "Number_1"

    assert slugify("a", valid_variable_name=False) == "a"
    assert slugify("1a", valid_variable_name=False) == "1a"
    assert slugify("1", valid_variable_name=False) == "1"


def test_parse_quantity():
    # From the metadata specification, the quantity is defined as
    # "name (units)" without backets in the name of the quantity
    assert parse_quantity("a (b)") == ("a", "b")
    assert parse_quantity("a (b/(c))") == ("a", "b/(c)")
    assert parse_quantity("a (c) (b/(c))") == ("a (c)", "b/(c)")
    assert parse_quantity("a [b]") == ("a [b]", "")
    assert parse_quantity("a [b]", opening="[", closing="]") == ("a", "b")


def test_is_hyperspy_signal():
    s = signals.Signal1D(np.zeros((5, 5, 5)))
    p = object()
    assert is_hyperspy_signal(s) is True
    assert is_hyperspy_signal(p) is False


def test_strlist2enumeration():
    assert strlist2enumeration([]) == ""
    assert strlist2enumeration("a") == "a"
    assert strlist2enumeration(["a"]) == "a"
    assert strlist2enumeration(["a", "b"]) == "a and b"
    assert strlist2enumeration(["a", "b", "c"]) == "a, b and c"



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest
import scipy.linalg

from hyperspy.exceptions import VisibleDeprecationWarning
from hyperspy.learn.rpca import orpca, rpca_godec
from hyperspy.signals import Signal1D


def compare_norms(a, b, tol=5e-3):
    assert a.shape == b.shape

    m, n = a.shape
    tol *= m * n
    n1 = np.linalg.norm(a)
    n2 = np.linalg.norm(b)

    assert np.linalg.norm((a / n1) - (b / n2)) < tol


class TestRPCA:
    def setup_method(self, method):
        # Define shape etc.
        m = 128  # Dimensionality
        n = 256  # Number of samples
        r = 3
        s = 0.01

        # Low-rank, sparse and noise matrices
        rng = np.random.RandomState(101)
        U = scipy.linalg.orth(rng.randn(m, r))
        V = rng.randn(n, r)
        A = U @ V.T
        E = 10 * rng.binomial(1, s, (m, n))
        G = 0.005 * rng.randn(m, n)
        X = A + E + G

        self.m = m
        self.n = n
        self.rank = r
        self.A = A
        self.X = X

    def test_default(self):
        X, E, U, S, V = rpca_godec(self.X, rank=self.rank)
        compare_norms(X, self.A)

    @pytest.mark.parametrize("power", [0, 1, 2])
    @pytest.mark.parametrize("maxiter", [1e2, 1e3])
    @pytest.mark.parametrize("tol", [1e-1, 1e-4])
    def test_tol_iter(self, power, maxiter, tol):
        X, E, U, S, V = rpca_godec(
            self.X, rank=self.rank, power=power, maxiter=maxiter, tol=tol
        )
        compare_norms(X, self.A)

    def test_regularization(self):
        X, E, U, S, V = rpca_godec(self.X, rank=self.rank, lambda1=0.01)
        compare_norms(X, self.A)

    @pytest.mark.parametrize("poisson", [True, False])
    def test_signal(self, poisson):
        # Note that s1.decomposition() operates on the transpose
        # i.e. (n_samples, n_features).
        x = self.X.copy().T.reshape(16, 16, 128)

        if poisson:
            x -= x.min()
            x[x <= 0] = 1e-16

        s1 = Signal1D(x)

        X, E = s1.decomposition(
            normalize_poissonian_noise=poisson,
            algorithm="RPCA",
            output_dimension=self.rank,
            return_info=True,
        )
        compare_norms(X, self.A.T)


class TestORPCA:
    def setup_method(self, method):
        # Define shape etc.
        m = 128  # Dimensionality
        n = 256  # Number of samples
        r = 3
        s = 0.01

        # Low-rank and sparse error matrices
        rng = np.random.RandomState(101)
        U = scipy.linalg.orth(rng.randn(m, r))
        V = rng.randn(n, r)
        A = U @ V.T
        E = 10 * rng.binomial(1, s, (m, n))
        X = A + E

        self.m = m
        self.n = n
        self.rank = r
        self.A = A
        self.X = X
        self.U = U

    def test_default(self):
        X, E, U, S, V = orpca(self.X, rank=self.rank, store_error=True)
        compare_norms(X, self.A)

    def test_project(self):
        L, R = orpca(self.X, rank=self.rank, project=True)

        assert L.shape == (self.m, self.rank)
        assert R.shape == (self.rank, self.n)

    def test_batch_size(self):
        L, R = orpca(self.X, rank=self.rank, batch_size=2)

        assert L.shape == (self.m, self.rank)
        assert R.shape == (self.rank, self.n)

    def test_method_BCD(self):
        X, E, U, S, V = orpca(self.X, rank=self.rank, store_error=True, method="BCD")
        compare_norms(X, self.A)

    @pytest.mark.parametrize("subspace_learning_rate", [1.0, 1.1])
    def test_method_SGD(self, subspace_learning_rate):
        X, E, U, S, V = orpca(
            self.X,
            rank=self.rank,
            store_error=True,
            method="SGD",
            subspace_learning_rate=subspace_learning_rate,
        )
        compare_norms(X, self.A)

    @pytest.mark.parametrize("subspace_momentum", [0.5, 0.1])
    def test_method_MomentumSGD(self, subspace_momentum):
        X, E, U, S, V = orpca(
            self.X,
            rank=self.rank,
            store_error=True,
            method="MomentumSGD",
            subspace_learning_rate=1.1,
            subspace_momentum=subspace_momentum,
        )
        compare_norms(X, self.A)

        with pytest.raises(ValueError, match=f"must be a float between 0 and 1"):
            _ = orpca(
                self.X, rank=self.rank, method="MomentumSGD", subspace_momentum=1.9
            )

    def test_init_rand(self):
        X, E, U, S, V = orpca(self.X, rank=self.rank, store_error=True, init="rand")
        compare_norms(X, self.A)

    def test_init_mat(self):
        X, E, U, S, V = orpca(self.X, rank=self.rank, store_error=True, init=self.U)
        compare_norms(X, self.A)

        with pytest.raises(ValueError, match=f"has to be a two-dimensional matrix"):
            mat = np.zeros(self.m)
            _ = orpca(self.X, rank=self.rank, init=mat)

        with pytest.raises(ValueError, match=f"has to be of shape"):
            mat = np.zeros((self.m, self.rank - 1))
            _ = orpca(self.X, rank=self.rank, init=mat)

    @pytest.mark.parametrize("rank", [3, 11])
    @pytest.mark.parametrize("training_samples", [16, 32])
    def test_training(self, rank, training_samples):
        X, E, U, S, V = orpca(
            self.X,
            rank=rank,
            store_error=True,
            init="qr",
            training_samples=training_samples,
        )
        compare_norms(X, self.A)

        with pytest.raises(ValueError, match=f"must be >="):
            _ = orpca(self.X, rank=self.rank, init="qr", training_samples=self.rank - 1)

    def test_regularization(self):
        X, E, U, S, V = orpca(
            self.X, rank=self.rank, store_error=True, lambda1=0.01, lambda2=0.02,
        )
        compare_norms(X, self.A)

    def test_exception_method(self):
        with pytest.raises(ValueError, match=f"'method' not recognised"):
            _ = orpca(self.X, rank=self.rank, method="uniform")

    def test_exception_init(self):
        with pytest.raises(ValueError, match=f"'init' not recognised"):
            _ = orpca(self.X, rank=self.rank, init="uniform")

    def test_warnings(self):
        with pytest.warns(
            VisibleDeprecationWarning,
            match=f"The argument `learning_rate` has been deprecated",
        ):
            _ = orpca(self.X, rank=self.rank, learning_rate=0.1)

        with pytest.warns(
            VisibleDeprecationWarning,
            match=f"The argument `momentum` has been deprecated",
        ):
            _ = orpca(self.X, rank=self.rank, momentum=0.1)

    @pytest.mark.parametrize("poisson", [True, False])
    def test_signal(self, poisson):
        # Note that s1.decomposition() operates on the transpose
        # i.e. (n_samples, n_features).
        x = self.X.copy().T.reshape(16, 16, 128)

        if poisson:
            x -= x.min()
            x[x <= 0] = 1e-16

        s1 = Signal1D(x)

        X, E = s1.decomposition(
            normalize_poissonian_noise=poisson,
            algorithm="ORPCA",
            output_dimension=self.rank,
            return_info=True,
        )
        compare_norms(X, self.A.T)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <http://www.gnu.org/licenses/>.

import math
from distutils.version import LooseVersion

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize, LogNorm, SymLogNorm, PowerNorm
from traits.api import Undefined
import logging
import inspect
import copy

from hyperspy.drawing import widgets
from hyperspy.drawing import utils
from hyperspy.signal_tools import ImageContrastEditor
from hyperspy.misc import math_tools
from hyperspy.misc import rgb_tools
from hyperspy.drawing.figure import BlittedFigure
from hyperspy.ui_registry import DISPLAY_DT, TOOLKIT_DT
from hyperspy.docstrings.plot import PLOT2D_DOCSTRING
from hyperspy.misc.test_utils import ignore_warning
from hyperspy.defaults_parser import preferences


_logger = logging.getLogger(__name__)


class ImagePlot(BlittedFigure):

    """Class to plot an image with the necessary machinery to update
    the image when the coordinates of an AxesManager change.

    Attributes
    ----------
    data_fuction : function or method
        A function that returns a 2D array when called without any
        arguments.
    %s
    pixel_units : {None, string}
        The pixel units for the scale bar.
    plot_indices : bool
    title : str
        The title is printed at the top of the image.

    """ % PLOT2D_DOCSTRING

    def __init__(self, title=""):
        super(ImagePlot, self).__init__()
        self.data_function = None
        self.data_function_kwargs = {}

        # Attribute matching the arguments of
        # `hyperspy._signal.signal2d.signal2D.plot`
        self.autoscale = "v"
        self.saturated_pixels = None
        self.norm = "auto"
        self.vmin = None
        self.vmax = None
        self.gamma = 1.0
        self.linthresh = 0.01
        self.linscale = 0.1
        self.scalebar = True
        self.scalebar_color = "white"
        self.axes_ticks = None
        self.axes_off = False
        self.axes_manager = None
        self.no_nans = False
        self.colorbar = True
        self.centre_colormap = "auto"
        self.min_aspect = 0.1

        # Other attributes
        self.pixel_units = None
        self._colorbar = None
        self.quantity_label = ''
        self.figure = None
        self.ax = None
        self.title = title

        # user provided numeric values
        self._vmin_numeric = None
        self._vmax_numeric = None
        self._vmax_auto = None
        # user provided percentile values
        self._vmin_percentile = None
        self._vmax_percentile = None
        # default values used when the numeric and percentile are None
        self._vmin_default = f"{preferences.Plot.saturated_pixels / 2}th"
        self._vmax_default = f"{100 - preferences.Plot.saturated_pixels / 2}th"
        # use to store internally the numeric value of contrast
        self._vmin = None
        self._vmax = None

        self._ylabel = ''
        self._xlabel = ''
        self.plot_indices = True
        self._text = None
        self._text_position = (0, 1.05,)
        self._aspect = 1
        self._extent = None
        self.xaxis = None
        self.yaxis = None
        self.ax_markers = list()
        self._user_scalebar = None
        self._auto_scalebar = False
        self._user_axes_ticks = None
        self._auto_axes_ticks = True
        self._is_rgb = False

    @property
    def vmax(self):
        if self._vmax_numeric is not None:
            return self._vmax_numeric
        elif self._vmax_percentile is not None:
            return self._vmax_percentile
        else:
            return self._vmax_default

    @vmax.setter
    def vmax(self, vmax):
        if isinstance(vmax, str):
            self._vmax_percentile = vmax
            self._vmax_numeric = None
        elif isinstance(vmax, (int, float)):
            self._vmax_numeric = vmax
            self._vmax_percentile = None
        elif vmax is None:
            self._vmax_percentile = self._vmax_numeric = None
        else:
            raise TypeError("`vmax` must be a number or a string.")

    @property
    def vmin(self):
        if self._vmin_numeric is not None:
            return self._vmin_numeric
        elif self._vmin_percentile is not None:
            return self._vmin_percentile
        else:
            return self._vmin_default

    @vmin.setter
    def vmin(self, vmin):
        if isinstance(vmin, str):
            self._vmin_percentile = vmin
            self._vmin_numeric = None
        elif isinstance(vmin, (int, float)):
            self._vmin_numeric = vmin
            self._vmin_percentile = None
        elif vmin is None:
            self._vmin_percentile = self._vmin_numeric = None
        else:
            raise TypeError("`vmax` must be a number or a string.")

    @property
    def axes_ticks(self):
        if self._user_axes_ticks is None:
            if self.scalebar is False:
                return True
            else:
                return self._auto_axes_ticks
        else:
            return self._user_axes_ticks

    @axes_ticks.setter
    def axes_ticks(self, value):
        self._user_axes_ticks = value

    @property
    def scalebar(self):
        if self._user_scalebar is None:
            return self._auto_scalebar
        else:
            return self._user_scalebar

    @scalebar.setter
    def scalebar(self, value):
        if value is False:
            self._user_scalebar = value
        else:
            self._user_scalebar = None

    def configure(self):
        xaxis = self.xaxis
        yaxis = self.yaxis

        if (xaxis.units == yaxis.units) and (xaxis.scale == yaxis.scale):
            self._auto_scalebar = True
            self._auto_axes_ticks = False
            self.pixel_units = xaxis.units
        else:
            self._auto_scalebar = False
            self._auto_axes_ticks = True

        # Signal2D labels
        self._xlabel = '{}'.format(xaxis)
        if xaxis.units is not Undefined:
            self._xlabel += ' ({})'.format(xaxis.units)

        self._ylabel = '{}'.format(yaxis)
        if yaxis.units is not Undefined:
            self._ylabel += ' ({})'.format(yaxis.units)

        # Calibrate the axes of the navigator image
        self._extent = [xaxis.axis[0] - xaxis.scale / 2.,
                        xaxis.axis[-1] + xaxis.scale / 2.,
                        yaxis.axis[-1] + yaxis.scale / 2.,
                        yaxis.axis[0] - yaxis.scale / 2.]
        self._calculate_aspect()
        if self.saturated_pixels is not None:
            from hyperspy.exceptions import VisibleDeprecationWarning
            VisibleDeprecationWarning("`saturated_pixels` is deprecated and will be "
                            "removed in 2.0. Please use `vmin` and `vmax` "
                            "instead.")
            self._vmin_percentile = self.saturated_pixels / 2
            self._vmax_percentile = self.saturated_pixels / 2


    def _calculate_aspect(self):
        xaxis = self.xaxis
        yaxis = self.yaxis
        factor = 1
        # Apply aspect ratio constraint
        if self.min_aspect:
            min_asp = self.min_aspect
            if yaxis.size / xaxis.size < min_asp:
                factor = min_asp * xaxis.size / yaxis.size
                self._auto_scalebar = False
                self._auto_axes_ticks = True
            elif yaxis.size / xaxis.size > min_asp ** -1:
                factor = min_asp ** -1 * xaxis.size / yaxis.size
                self._auto_scalebar = False
                self._auto_axes_ticks = True
        self._aspect = np.abs(factor * xaxis.scale / yaxis.scale)

    def _calculate_vmin_max(self, data, auto_contrast=False,
                            vmin=None, vmax=None):
        # Calculate vmin and vmax using `utils.contrast_stretching` when:
        # - auto_contrast is True
        # - self.vmin or self.vmax is of tpye str
        if (auto_contrast and (isinstance(self.vmin, str) or
            isinstance(self.vmax, str))):
            with ignore_warning(category=RuntimeWarning):
                # In case of "All-NaN slices"
                vmin, vmax = utils.contrast_stretching(
                    data, self.vmin, self.vmax)
        else:
            vmin, vmax = self._vmin_numeric, self._vmax_numeric
        # provided vmin, vmax override the calculated value
        if isinstance(vmin, (int, float)):
            vmin = vmin
        if isinstance(vmax, (int, float)):
            vmax = vmax
        if vmin == np.nan:
            vmin = None
        if vmax == np.nan:
            vmax = None

        return vmin, vmax

    def create_figure(self, max_size=None, min_size=2, **kwargs):
        """Create matplotlib figure

        The figure size is automatically computed by default, taking into
        account the x and y dimensions of the image. Alternatively the figure
        size can be defined by passing the ``figsize`` keyword argument.

        Parameters
        ----------
        max_size, min_size: number
            The maximum and minimum size of the axes in inches. These have
            no effect when passing the ``figsize`` keyword to manually set
            the figure size.
        **kwargs
            All keyword arguments are passed to
            :py:func:`matplotlib.pyplot.figure`.

        """
        if "figsize" not in kwargs:
            if self.scalebar is True:
                wfactor = 1.0 + plt.rcParams['font.size'] / 100
            else:
                wfactor = 1

            height = abs(self._extent[3] - self._extent[2]) * self._aspect
            width = abs(self._extent[1] - self._extent[0])
            figsize = np.array((width * wfactor, height)) * \
                max(plt.rcParams['figure.figsize']) / \
                max(width * wfactor, height)
            kwargs["figsize"] = figsize.clip(min_size, max_size)
        if "disable_xyscale_keys" not in kwargs:
            kwargs["disable_xyscale_keys"] = True
        super().create_figure(**kwargs)

    def create_axis(self):
        self.ax = self.figure.add_subplot(111)
        self.ax.set_title(self.title)
        self.ax.set_xlabel(self._xlabel)
        self.ax.set_ylabel(self._ylabel)
        if self.axes_ticks is False:
            self.ax.set_xticks([])
            self.ax.set_yticks([])
        self.ax.hspy_fig = self
        if self.axes_off:
            self.ax.axis('off')

    def plot(self, data_function_kwargs={}, **kwargs):
        self.data_function_kwargs = data_function_kwargs
        self.configure()
        if self.figure is None:
            self.create_figure()
            self.create_axis()

        if (not self.axes_manager or self.axes_manager.navigation_size == 0):
            self.plot_indices = False
        if self.plot_indices is True:
            if self._text is not None:
                self._text.remove()
            self._text = self.ax.text(
                *self._text_position,
                s=str(self.axes_manager.indices),
                transform=self.ax.transAxes,
                fontsize=12,
                color='red',
                animated=self.figure.canvas.supports_blit)
        for marker in self.ax_markers:
            marker.plot()
        for attribute in ['vmin', 'vmax']:
            if attribute in kwargs.keys():
                setattr(self, attribute, kwargs.pop(attribute))
        self.update(data_changed=True, auto_contrast=True, **kwargs)
        if self.scalebar is True:
            if self.pixel_units is not None:
                self.ax.scalebar = widgets.ScaleBar(
                    ax=self.ax,
                    units=self.pixel_units,
                    animated=self.figure.canvas.supports_blit,
                    color=self.scalebar_color,
                )

        if self.colorbar:
            self._add_colorbar()

        if hasattr(self.figure, 'tight_layout'):
            try:
                if self.axes_ticks == 'off' and not self.colorbar:
                    plt.subplots_adjust(0, 0, 1, 1)
                else:
                    self.figure.tight_layout()
            except BaseException:
                # tight_layout is a bit brittle, we do this just in case it
                # complains
                pass

        self.connect()
        self.figure.canvas.draw()

    def _add_colorbar(self):
        # Bug extend='min' or extend='both' and power law norm
        # Use it when it is fixed in matplotlib
        self._colorbar = plt.colorbar(self.ax.images[0], ax=self.ax)
        self.set_quantity_label()
        self._colorbar.set_label(
            self.quantity_label, rotation=-90, va='bottom')
        self._colorbar.ax.yaxis.set_animated(
            self.figure.canvas.supports_blit)

    def _update_data(self):
        # self._current_data caches the displayed data.
        self._current_data =  self.data_function(
                axes_manager=self.axes_manager,
                **self.data_function_kwargs)

    def update(self, data_changed=True, auto_contrast=None, vmin=None,
               vmax=None, **kwargs):
        """
        Parameters
        ----------
        data_changed : bool, optional
            Fetch and update the data to display. It can be used to avoid
            unnecessarily reading of the data from disk with working with lazy
            signal. The default is True.
        auto_contrast : bool or None, optional
            Force automatic resetting of the intensity limits. If None, the
            intensity values will change when 'v' is in autoscale.
            Default is None.
        vmin, vmax : float or str
            `vmin` and `vmax` are used to normalise the displayed data.
        **kwargs : dict
            The kwargs are passed to :py:func:`matplotlib.pyplot.imshow`.

        Raises
        ------
        ValueError
            When the selected ``norm`` is not valid or the data are not
            compatible with the selected ``norm``.
        """
        if auto_contrast is None:
            auto_contrast = 'v' in self.autoscale
        if data_changed:
            # When working with lazy signals the following may reread the data
            # from disk unnecessarily, for example when updating the image just
            # to recompute the histogram to adjust the contrast. In those cases
            # use `data_changed=True`.
            _logger.debug("Updating image slowly because `data_changed=True`")
            self._update_data()
        data = self._current_data
        if rgb_tools.is_rgbx(data):
            self.colorbar = False
            data = rgb_tools.rgbx2regular_array(data, plot_friendly=True)
            data = self._current_data = data
            self._is_rgb = True
        ims = self.ax.images

        # Turn on centre_colormap if a diverging colormap is used.
        if not self._is_rgb and self.centre_colormap == "auto":
            if "cmap" in kwargs:
                cmap = kwargs["cmap"]
            elif ims:
                cmap = ims[0].get_cmap().name
            else:
                cmap = plt.cm.get_cmap().name
            if cmap in utils.MPL_DIVERGING_COLORMAPS:
                self.centre_colormap = True
            else:
                self.centre_colormap = False
        redraw_colorbar = False


        for marker in self.ax_markers:
            marker.update()

        if not self._is_rgb:
            def format_coord(x, y):
                try:
                    col = self.xaxis.value2index(x)
                except ValueError:  # out of axes limits
                    col = -1
                try:
                    row = self.yaxis.value2index(y)
                except ValueError:
                    row = -1
                if col >= 0 and row >= 0:
                    z = data[row, col]
                    if np.isfinite(z):
                        return f'x={x:1.4g}, y={y:1.4g}, intensity={z:1.4g}'
                return f'x={x:1.4g}, y={y:1.4g}'
            self.ax.format_coord = format_coord

            old_vmin, old_vmax = self._vmin, self._vmax

            if auto_contrast:
                vmin, vmax = self._calculate_vmin_max(data, auto_contrast,
                                                      vmin, vmax)
            else:
                # use the value store internally when not explicitely defined
                if vmin is None:
                    vmin = old_vmin
                if vmax is None:
                    vmax = old_vmax

            # If there is an image, any of the contrast bounds have changed and
            # the new contrast bounds are not the same redraw the colorbar.
            if (ims and (old_vmin != vmin or old_vmax != vmax) and
                    vmin != vmax):
                redraw_colorbar = True
                ims[0].autoscale()
            if self.centre_colormap:
                vmin, vmax = utils.centre_colormap_values(vmin, vmax)

            if self.norm == 'auto' and self.gamma != 1.0:
                self.norm = 'power'
            norm = copy.copy(self.norm)
            if norm == 'power':
                # with auto norm, we use the power norm when gamma differs from its
                # default value.
                norm = PowerNorm(self.gamma, vmin=vmin, vmax=vmax)
            elif norm == 'log':
                if np.nanmax(data) <= 0:
                    raise ValueError('All displayed data are <= 0 and can not '
                                    'be plotted using `norm="log"`. '
                                    'Use `norm="symlog"` to plot on a log scale.')
                if np.nanmin(data) <= 0:
                    vmin = np.nanmin(np.where(data > 0, data, np.inf))

                norm = LogNorm(vmin=vmin, vmax=vmax)
            elif norm == 'symlog':
                sym_log_kwargs = {'linthresh':self.linthresh,
                                  'linscale':self.linscale,
                                  'vmin':vmin, 'vmax':vmax}
                if LooseVersion(matplotlib.__version__) >= LooseVersion("3.2"):
                    sym_log_kwargs['base'] = 10
                norm = SymLogNorm(**sym_log_kwargs)
            elif inspect.isclass(norm) and issubclass(norm, Normalize):
                norm = norm(vmin=vmin, vmax=vmax)
            elif norm not in ['auto', 'linear']:
                raise ValueError("`norm` paramater should be 'auto', 'linear', "
                                "'log', 'symlog' or a matplotlib Normalize  "
                                "instance or subclass.")
            else:
                # set back to matplotlib default
                norm = None

            self._vmin, self._vmax = vmin, vmax

        redraw_colorbar = redraw_colorbar and self.colorbar

        if self.plot_indices is True:
            self._text.set_text(self.axes_manager.indices)
        if self.no_nans:
            data = np.nan_to_num(data)

        if ims:  # the images has already been drawn previously
            ims[0].set_data(data)
            # update extent:
            if 'x' in self.autoscale:
                self._extent[0] = self.xaxis.axis[0] - self.xaxis.scale / 2
                self._extent[1] = self.xaxis.axis[-1] + self.xaxis.scale / 2
                self.ax.set_xlim(self._extent[:2])
            if 'y' in self.autoscale:
                self._extent[2] = self.yaxis.axis[-1] + self.yaxis.scale / 2
                self._extent[3] = self.yaxis.axis[0] - self.yaxis.scale / 2
                self.ax.set_ylim(self._extent[2:])
            if 'x' in self.autoscale or 'y' in self.autoscale:
                ims[0].set_extent(self._extent)
            self._calculate_aspect()
            self.ax.set_aspect(self._aspect)
            if not self._is_rgb:
                ims[0].set_norm(norm)
                ims[0].norm.vmax, ims[0].norm.vmin = vmax, vmin
            if redraw_colorbar:
                self._colorbar.draw_all()
                self._colorbar.solids.set_animated(
                    self.figure.canvas.supports_blit
                )
            else:
                ims[0].changed()
            if self.figure.canvas.supports_blit:
                self._update_animated()
            else:
                self.figure.canvas.draw_idle()
        else:  # no signal have been drawn yet
            new_args = {'interpolation': 'nearest',
                        'extent': self._extent,
                        'aspect': self._aspect,
                        'animated': self.figure.canvas.supports_blit,
                        }
            if not self._is_rgb:
                if norm is None:
                    new_args.update({'vmin': vmin, 'vmax':vmax})
                else:
                    new_args['norm'] = norm
            new_args.update(kwargs)
            self.ax.imshow(data, **new_args)
            self.figure.canvas.draw_idle()

        if self.axes_ticks == 'off':
            self.ax.set_axis_off()

    def _update(self):
        # This "wrapper" because on_trait_change fiddles with the
        # method arguments and auto contrast does not work then
        self.update(data_changed=False)

    def gui_adjust_contrast(self, display=True, toolkit=None):
        if self._is_rgb:
            raise NotImplementedError(
                "Constrast adjustment of RGB images is not implemented")
        ceditor = ImageContrastEditor(self)
        return ceditor.gui(display=display, toolkit=toolkit)
    gui_adjust_contrast.__doc__ = \
        """
        Display widgets to adjust image contrast if available.

        Parameters
        ----------
        %s
        %s

        """ % (DISPLAY_DT, TOOLKIT_DT)

    def connect(self):
        self.figure.canvas.mpl_connect('key_press_event',
                                        self.on_key_press)
        if self.axes_manager:
            if self.update not in self.axes_manager.events.indices_changed.connected:
                self.axes_manager.events.indices_changed.connect(self.update, [])
            if self.disconnect not in self.events.closed.connected:
                self.events.closed.connect(self.disconnect, [])

    def disconnect(self):
        if self.axes_manager:
            if self.update in self.axes_manager.events.indices_changed.connected:
                self.axes_manager.events.indices_changed.disconnect(self.update)

    def on_key_press(self, event):
        if event.key == 'h':
            self.gui_adjust_contrast()
        if event.key == 'l':
            self.toggle_norm()

    def toggle_norm(self):
        self.norm = 'linear' if self.norm == 'log' else 'log'
        self.update(data_changed=False)
        if self.colorbar:
            self._colorbar.remove()
            self._add_colorbar()
            self.figure.canvas.draw_idle()

    def set_quantity_label(self):
        if 'power_spectrum' in self.data_function_kwargs.keys():
            if self.norm == 'log':
                if 'FFT' in self.quantity_label:
                    self.quantity_label = self.quantity_label.replace(
                        'Power spectral density', 'FFT')
                else:
                    of = ' of ' if self.quantity_label else ''
                    self.quantity_label = 'Power spectral density' + of + \
                        self.quantity_label
            else:
                self.quantity_label = self.quantity_label.replace(
                    'Power spectral density of ', '')
                self.quantity_label = self.quantity_label.replace(
                    'Power spectral density', '')

    def set_contrast(self, vmin, vmax):
        self.vmin, self.vmax = vmin, vmax
        self.update(data_changed=False, auto_contrast=True)

    def optimize_colorbar(self,
                          number_of_ticks=5,
                          tolerance=5,
                          step_prec_max=1):
        vmin, vmax = self.vmin, self.vmax
        _range = vmax - vmin
        step = _range / (number_of_ticks - 1)
        step_oom = math_tools.order_of_magnitude(step)

        def optimize_for_oom(oom):
            self.colorbar_step = math.floor(step / 10 ** oom) * 10 ** oom
            self.colorbar_vmin = math.floor(vmin / 10 ** oom) * 10 ** oom
            self.colorbar_vmax = self.colorbar_vmin + \
                self.colorbar_step * (number_of_ticks - 1)
            self.colorbar_locs = (
                np.arange(0, number_of_ticks) *
                self.colorbar_step +
                self.colorbar_vmin)

        def check_tolerance():
            if abs(self.colorbar_vmax - vmax) / vmax > (
                tolerance / 100.) or abs(self.colorbar_vmin - vmin
                                         ) > (tolerance / 100.):
                return True
            else:
                return False

        optimize_for_oom(step_oom)
        i = 1
        while check_tolerance() and i <= step_prec_max:
            optimize_for_oom(step_oom - i)
            i += 1
